
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// DOCBDepc.h: DOCB dependency protocol (classes)
///////////////////////////////////////////////////////////////////////////
enum dep_t {
  random_dep, sref_dep, dref_dep, trav_dep, class_dep
};

class RandomDep {
  
 private:

  set<int> * _setObjs;
  multimap<const int, int, int_cmp> * _classObjMap;
  vector<int> _hotObjects;
  vector<int> _coldObjects;
  float _hotRegionSize;
  float _hotRegionProb;
  objAssMeth_t _assignMethod;
  
 public:
  
  RandomDep(float hotRegionSize, float hotRegionProb, 
	    objAssMeth_t assignMethod);
  void AssignObjects();
  void InsertObject(int classId, int objId);
  int NextRoot();
  set<int> * GetCandidateSet();
  ~RandomDep();
};

class ByRefDep {
  
 private:

  hash_map<int, set<int> * , hash<int>, eqoid> _refs;

 public:

  ByRefDep();
  void InsertRef(int rootId, int refId);
  // by default the next root selected is via random selection of objects
  // in the reference of the current root.
  int NextRoot(int currRoot);
  set<int> * GetCandidateSet(int rootId);

  // returns all the roots that are stored in this object.
  vector<int> GetRootSet();

  ~ByRefDep();
};


class TraversedDep {
  
 private:
  
  multimap<const int, int, int_cmp> _traversedObjects;
  int _currRoot;
  float _C;

 public:
  
  TraversedDep(float C);  
  
  // The rootId arguement is just used by the class to reset the 
  // traversedObjects set when a new root is encountered.
  void InsertObject(int rootId, int traversedId);

  // The currRoot arguement is just used to check if the set of traversed 
  // objects stored are for the current root.
  int NextRoot(int currRoot);
  vector<int> GetCandidateSet(int currRoot);
};

class SameClassDep {
  
 private:
  
  // The set of candidate objects for a root object, is stored in a continuous
  // range starting from the index indicated in _setStartIndexMap, if the end
  // of the range goes over the edge % is used to wrap it around.
  hash_map<int, int, hash<int>, eqoid> _setStartIndexMap;
  multimap<int, int, int_cmp> _classToObjMap;
  hash_map<int, int, hash<int>, eqoid> _classSizeMap;
  hash_map<int, int, hash<int>, eqoid> _objToClassMap;
  float _U;

 public:

  SameClassDep(float U);
  void InsertObject(int classId, int objId);

  // assigns a set of same class objects as the candidate set for each object
  void AssignSets();
  int NextRoot(int currRoot);
  vector<int> GetCandidateSet(int currRoot);

  // returns all the roots that are stored in this object.
  vector<int> GetRootSet();

};


class HybridDep {
  
 private:
  
  float _depRandomProb;
  float _depSRefProb;
  float _depDRefProb;
  float _depTraversedProb;
  float _depClassProb;
  
  RandomDep * _randomDep;
  ByRefDep * _bySRefDep;
  ByRefDep * _byDRefDep;
  TraversedDep * _traversedDep;
  SameClassDep * _sameClassDep;
  int _R;
  int _numRoots;

 public:

  HybridDep(float depRandomProb, float depSRefProb, float depDRefProb,
	    float depTraversedProb, float depClassProb, float U, float C,
	    float randHotRegSize, float randHotRegionProb,
	    objAssMeth_t randAssMethod, int R);

  void InsertObjectRandom(int classId, int objId);
  void InsertSRef(int rootId, int refId);
  void InsertDRef(int rootId, int refId);
  void InsertTraversedObject(int rootId, int traversedId);
  void InsertSameClassObject(int classId, int objId);

  
  // should be called after both randomDep and sameClassDep has all there
  // objects inserted.
  void Assign();
  
  int NextRoot(int currRoot);

  ~HybridDep();

};
