////////////////////////////////////////////////////////////////////
// Package SIMULation (C++ discrete event simulation package)
// Version 1.1 g++, mars 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// Updated
// -------
// Name: Zhen He
// Institution: Australian National University
// Date: Feb. 2002
/////////////////////////////////////////////////////////////////////
// eventc.h : Definition des methodes du gestionnaire d'evenements
// Variable selon les systemes simules
// Implantation VOODB2 (avril 1998)
/////////////////////////////////////////////////////////////////////
// Notice: all opportunistic writes are placed before reads, this
//         is since opportunistic writes automatically imply opportunistic
//         reads.


// Prefetch pseudo code:
//
// PRECONDITION of prefetching: no large objects.
//
// 1) At time of prefetch, record how long it took to perform the prefetch.
//
// 2) Mark page as one just been prefetched.
//
// 3) When the prefetched page actually get used or the next I/O is generated
//    (for any other page), compare curretn timestamp with the prefetch time
//    stamp (where prefetch time stamp is when the prefetch completed, without
//    incrementing current time).  Current time should be set to the max of 
//    the two timestamps.
//
// 4) If the prefetch has not been performed then wait for prefetch to complete
//    before proceeding with the next I/O \footnote{This means earlier prefetch
//    requests get higher priority when comopared to new demand requests.
//    (need to consider if we should generate new prefetch requests when 
//    previous prefetch requests are still pending.)} or using the prefetched
//    page.  - update the current time when prefetch complete.


#include "stdlib.h"
#include "string.h"
#include <hash_set.h>
#include <stack.h>

int currentTransactionRoot;
Client * start_client;

vector<int> root_vect;

int min(int a, int b) {
  if (a < b) {
    return a;
  }
  else {
    return b;
  }
}

hash_map<int, DROPageStats *, hash<int>, eqoid> DROPageStatsMap;

/////////////////////////////////////////////////////////////////////
// Class Client
/////////////////////////////////////////////////////////////////////

// Class Client : Constructor

Client::Client() {
  reached_end = false;
  toaccess = NULL;
  update_root_obj = false;
}


/////////////////////////////////////////////////////////////////////
//   ASSE EventManager
/////////////////////////////////////////////////////////////////////

// CLASSE EventManager : Constructeur

EventManager::EventManager(Simulation *sim) {

  int i;
  char name[STRS], num[STRS];

  simul=sim;

  // Instanciation des resources passives

  // Processeur serveur
  serverpr=new Resource("Server Processor",1,simul);

  // Processeurs clients
  if (SYSCLASS==1) // Architecture centralisee
    // Pas de client, tous les traitements sont effectues sur la meme machine
    for (i=0; i<NUSERS; i++) clientpr[i]=serverpr;
  else if ((SYSCLASS>=2) && (SYSCLASS<=4))
    for (i=0; i<NUSERS; i++) {
      strcpy(name,"Client #");
      // itoa(i,num,10);
      sprintf(num,"%d",i);
      strcat(name,num);
      strcat(name," Processor");
      clientpr[i]=new Resource(name,1,simul);
    }
  else if (SYSCLASS==9) {  // Architecture utilisateur
    // Code a inserer ici
  }

  // Disque serveur
  disk=new Resource("Disk",1,simul);

  // Base de donnees
  db=new Resource("Database",MULTILVL,simul);

  // Instanciation des resources actives

  // Clients
  for (i=0; i<NUSERS; i++) {
    strcpy(name,"User #");
    //itoa(i,num,10);
    sprintf(num,"%d",i);
    strcat(name,num);
    user[i]=new User(name,1,simul,i,clientpr[i]);
  }

  // Gestionnaire de transactions
  if (SYSCLASS==4)              // Serveur de BD
    transmgr=new TransMgr("Transaction Mgr",INFINITE,simul,serverpr,db,1/NETTHRU);
  else if (SYSCLASS<=3)
    transmgr=new TransMgr("Transaction Mgr",INFINITE,simul,clientpr[0],db,0);
  else {                        // Architecture utilisateur
    // Code a inserer ici
  }

  // Gestionnaire de groupement
  if (SYSCLASS==4)              // Serveur de BD
    clustmgr=new ClustMgr("Clustering Mgr",INFINITE,simul,serverpr,db);
  else if (SYSCLASS<=3)
    clustmgr=new ClustMgr("Clustering Mgr",INFINITE,simul,clientpr[0],db);
  else {                        // Architecture utilisateur
    // Code a inserer ici
  }

  // Gestionnaire d'objets
  if ((SYSCLASS==1) || (SYSCLASS==4))   // Systeme centralise ou sBD
    objmgr=new ObjMgr("Object Mgr",INFINITE,simul,serverpr,0);
  else if (SYSCLASS==2)                 // Serveur d'objets
    objmgr=new ObjMgr("Object Mgr",INFINITE,simul,serverpr,1/NETTHRU);
  else if (SYSCLASS==3)                 // Serveur de pages
    objmgr=new ObjMgr("Object Mgr",INFINITE,simul,clientpr[0],0);
  else {                                // Architecture utilisateur
    // Code a inserer ici
  }
  clustmgr->SetObjPtr(objmgr);

  // Gestionnaire de cache
  if (SYSCLASS==3)                      // Serveur de pages
    buffmgr=new BuffMgr("Buffering Mgr",INFINITE,simul,serverpr,
			PGSIZE/NETTHRU);
  else if ((SYSCLASS<=2) || (SYSCLASS==4))
    buffmgr=new BuffMgr("Buffering Mgr",INFINITE,simul,serverpr,0);
  else {                                // Architecture utilisateur
    // Code a inserer ici
  }
  clustmgr->SetBuffPtr(buffmgr);
  objmgr->SetBuffPtr(buffmgr);
  // Sous-systeme des Entrees/Sorties
  if (SYSCLASS!=9) {
    iosubsys=new IOSubSys("I/O Subsystem",INFINITE,simul,
			  serverpr,disk);
    iosubsys->SetBuffPtr(buffmgr);
  }
  else {                                // Architecture utilisateur
    // Code a inserer ici
  }
}

// CLASSE EventManager : Destructeur

EventManager::~EventManager() {

  int i;

  // Destruction des resources passives
  delete serverpr;
  for (i=0; i<NUSERS; i++) delete clientpr[i];
  delete disk;
  delete db;

  // Destruction des resources actives
  for (i=0; i<NUSERS; i++) delete user[i];
  delete transmgr;
  delete clustmgr;
  delete objmgr;
  delete buffmgr;
  delete iosubsys;
}

// CLASSE EventManager : Execution des evenements

void EventManager::ExecuteEvent(int code, Client *client) {

  int i;
  Client *newcli;

//  printf("Performing event #%d at time %f\n",code,simul->Tnow());

  switch(code) {

  // Evenement initial : creation d'un client par utilisateur

  case 0:{ client->cliid=0;
           client->type=0;
           user[client->cliid]->Event10(client);
           for (i=1; i<NUSERS; i++) {
             newcli=simul->NewClient();
             newcli->cliid=i;
             client->type=0;
             user[newcli->cliid]->Event10(newcli);
           }
           break;
         }

  // Evenements courants

  case 10: user[client->cliid]->Event10(client);break;
  case 11: user[client->cliid]->Event11(client);break;
  case 12: user[client->cliid]->Event12(client);break;
  case 13: user[client->cliid]->Event13(client);break;

  case 20: transmgr->Event20(client);break;
  case 21: transmgr->Event21(client);break;
  case 22: transmgr->Event22(client);break;
  case 23: transmgr->Event23(client);break;
  case 24: transmgr->Event24(client);break;
  case 25: transmgr->Event25(client);break;
  case 26: transmgr->Event26(client);break;
  case 27: transmgr->Event27(client);break;
  case 28: transmgr->Event28(client);break;

  case 30: clustmgr->Event30(client);break;
  case 31: clustmgr->Event31(client);break;
  case 32: clustmgr->Event32(client);break;
  case 33: clustmgr->Event33(client);break;
  case 34: clustmgr->Event34(client);break;
  case 35: clustmgr->Event35(client);break;
  case 36: clustmgr->Event36(client);break;
  case 37: clustmgr->Event37(client);break;
  case 38: clustmgr->Event38(client);break;

  case 40: objmgr->Event40(client);break;
  case 41: objmgr->Event41(client);break;
  case 42: objmgr->Event42(client);break;
  case 43: objmgr->Event43(client);break;

  case 50: buffmgr->Event50(client);break;
  case 51: buffmgr->Event51(client);break;
  case 52: buffmgr->Event52(client);break;
  case 53: buffmgr->Event53(client);break;
  case 54: buffmgr->Event54(client);break;
  case 55: buffmgr->Event55(client);break;

  case 60: iosubsys->Event60(client);break;
  case 61: iosubsys->Event61(client);break;
  case 62: iosubsys->Event62(client);break;
  case 63: iosubsys->Event63(client);break;

  case 100: transmgr->Event100(client);break;


  default: printf("Error: unknown event #%d at time %f\n",code,simul->Tnow());
  }
}

// CLASSE EventManager : Initialisation des statistiques pour chaque ressource

void EventManager::Init() {

  int i;

  // Ressources passives
  serverpr->ResetStats();
  for (i=0; i<NUSERS; i++) clientpr[i]->ResetStats();

  disk->ResetStats();

  db->ResetStats();


  // Ressources actives
  for (i=0; i<NUSERS; i++) user[i]->ResetStats();
  transmgr->ResetStats();
  clustmgr->ResetStats();
  objmgr->ResetStats();
  buffmgr->ResetStats();
  iosubsys->ResetStats();


  // Statistiques specifiques a VOODB2
  for (i=0; i<NTTRANS+1; i++) {
    NBTRANSC[i]=0;
    NBTRANSC2[i]=0;
    RTIMEC[i]=0;
    RTIMEC2[i]=0;
    NBOBJC[i]=0;
    NBOBJC2[i]=0;
  }
  NBCLC=0;
  NBCLC2=0;
  CTIMEC=0;
  CTIMEC2=0;
  TCTIMEC=0;
  TCTIMEC2=0;
  NBIOC=0;
  NBIOC2=0;
  TNBIOC=0;
  TNBIOC2=0;
  CNBIOC=0;
  CNBIOC2=0;
  WNBIOC=0;
  WNBIOC2=0;
  PNBIOC=0;
  PNBIOC2=0;
  NBPGC=0;
  NBPGC2=0;
  NCC=0;
  NCC2=0;
  TCC=0;
  TCC2=0;

}

// CLASSE EventManager : Initialisation des replications

void EventManager::InitRep() {

  int i;


  // Generation de la base et placement initial des objets
  
  STATS = new CStats();
  PQ = new PQueue();
  if ((OPPORTUNISTIC_READ == 1) || (OPPORTUNISTIC_WRITE == 1)) {
    buffmgr->StartIter(&STATS->GPBufferIter);
  }
  else {
    STATS->GPBufferIter = 0;
  }

  HOT_MOVE_FORWARD = true;
  HOT_START = 0.0;

  clustmgr->Initialise();

  DB=new Database();
  DB->Generation();
  DB->MoveDepAssignObjects();
  DB->Initp();

  // Initialise DRO stats
  vector <Page *> pageList;

  pageList = DB->GetPageList();

  Page * currPage;
  int currOid;
  int currPid;

  if (CLUSTP == 3) {

    for (unsigned int i = 0; i < pageList.size(); i++) {
      
      currPage = pageList[i];
      
      currPid = currPage->PageNum();
      DROPageStats * pageStats = new DROPageStats(currPid);
      
      for (int j = 1; j <= currPage->Objects()->NbElem(); j++) {
	currOid = currPage->Objects()->Scan(j);
	pageStats->insertObject(currOid);
      }
      
      DROPageStatsMap[currPid] = pageStats;
      
    }
  }

 
//  DB->Display();
//  DB->DisplayPages();

  // Echancier
  simul->Sched()->Purge();

  // Ressources passives
  serverpr->ResetCounters();
  serverpr->PurgeQueue();
  for (i=0; i<NUSERS; i++) {
    clientpr[i]->ResetCounters();
    clientpr[i]->PurgeQueue();
  }
  disk->ResetCounters();
  disk->PurgeQueue();
  db->ResetCounters();
  db->PurgeQueue();

  // Ressources actives
  for (i=0; i<NUSERS; i++) {
    user[i]->ResetCounters();
    user[i]->PurgeQueue();
  }
  transmgr->ResetCounters();
  transmgr->PurgeQueue();
  clustmgr->ResetCounters();
  clustmgr->PurgeQueue();
  objmgr->ResetCounters();
  objmgr->PurgeQueue();
  buffmgr->ResetCounters();
  buffmgr->PurgeQueue();
  buffmgr->InitBuffer();
  iosubsys->ResetCounters();
  iosubsys->PurgeQueue();

  // Statistiques specifiques a VOODB2
  for (i=0; i<NTTRANS+1; i++) {
    NBTRANS[i]=0;
    RTIME[i]=0;
    NBST[i]=0;
    NBOBJ[i]=0;
  }
  NBCL=0;
  CTIME=0;
  TCTIME=0;
  NBIO=0;
  TNBIO=0;
  CNBIO=0;
  WNBIO=0;
  NCLUSTERS=0;
  TCLUSTERS=0;
}

// CLASSE EventManager : Calcul des statistiques pour chaque ressource

void EventManager::Stats() {

  int i;
  float rtimeinc;

  // Ressources passives
  serverpr->Stats();
  for (i=0; i<NUSERS; i++) clientpr[i]->Stats();
  disk->Stats();
  db->Stats();

  // Ressources actives
  for (i=0; i<NUSERS; i++) user[i]->Stats();
  transmgr->Stats();
  clustmgr->Stats();
  objmgr->Stats();
  buffmgr->Stats();
  iosubsys->Stats();

  // Statistiques specifiques a VOODB2
  for (i=0; i<NTTRANS+1; i++) {
    NBTRANSC[i]+=NBTRANS[i];
    NBTRANSC2[i]+=NBTRANS[i]*NBTRANS[i];
    if (NBTRANS[i]!=0) rtimeinc=(RTIME[i]+(NBST[i]-NBTRANS[i])*simul->Tnow())/NBTRANS[i];
    else rtimeinc=0;
    RTIMEC[i]+=rtimeinc;
    RTIMEC2[i]+=rtimeinc*rtimeinc;
    if (NBTRANS[i]!=0) {
      NBOBJC[i]+=NBOBJ[i]/NBTRANS[i];
      NBOBJC2[i]+=(NBOBJ[i]/NBTRANS[i])*(NBOBJ[i]/NBTRANS[i]);
    } else {
      NBOBJC[i]=0;
      NBOBJC2[i]=0;
    }
  }
  NBCLC+=NBCL;
  NBCLC2+=NBCL*NBCL;
  CTIMEC+=CTIME;
  CTIMEC2+=CTIME*CTIME;
  TCTIMEC+=TCTIME;
  TCTIMEC2+=TCTIME*TCTIME;
  NBIOC+=NBIO;
  NBIOC2+=NBIO*NBIO;
  TNBIOC+=TNBIO;
  TNBIOC2+=TNBIO*TNBIO;
  CNBIOC+=CNBIO;
  CNBIOC2+=CNBIO*CNBIO;
  WNBIOC+=WNBIO;
  WNBIOC2+=WNBIO*WNBIO;
  PNBIOC+=PNBIO;
  PNBIOC2+=PNBIO*PNBIO;
  NBPGC+=DB->NbPages();
  NBPGC2+=DB->NbPages()*DB->NbPages();
  NCC+=NCLUSTERS;
  NCC2+=NCLUSTERS*NCLUSTERS;
  if (NCLUSTERS>0) {
    TCC+=TCLUSTERS/NCLUSTERS;
    TCC2+=(TCLUSTERS*TCLUSTERS)/(NCLUSTERS*NCLUSTERS);
  }

  _ltef_num_ideal_pages = transmgr->GetNumIdealPages();  


  // Destruction de la base de donnees en fin de replication
  delete DB;
  delete STATS;
  delete PQ;

}


// CLASSE EventManager : Affichage des statistiques pour chaque ressource

void EventManager::DisplayStats() {

  int i;

  printf("\n\n*** SIMULATION STATISTICS ***\n");

  //

  printf("\n*** PASSIVE RESOURCES\n");
  serverpr->DisplayStats();
  for (i=0; i<NUSERS; i++) clientpr[i]->DisplayStats();
  disk->DisplayStats();
  db->DisplayStats();

  //

  printf("\n*** ACTIVE RESOURCES\n");
  for (i=0; i<NUSERS; i++) user[i]->DisplayStats();
  transmgr->DisplayStats();
  clustmgr->DisplayStats();
  objmgr->DisplayStats();
  buffmgr->DisplayStats();
  iosubsys->DisplayStats();

  // Statistiques specifiques a VOODB2

  printf("\n*** TRANSACTIONS\n\n");

  //

  printf("Mean number of transactions\n\n");

  float nbtransm[NTTRANS+1], nbtransd[NTTRANS+1], nbtransi[NTTRANS+1];

  for (i=0; i<NTTRANS+1; i++) {
    nbtransm[i]=NBTRANSC[i]/NREPLIC;
    nbtransd[i]=(NREPLIC*NBTRANSC2[i]-NBTRANSC[i]*NBTRANSC[i])/(NREPLIC*NREPLIC);
    if (nbtransd[i]>0) nbtransd[i]=sqrt(nbtransd[i]);
    else nbtransd[i]=0;
    if (NREPLIC>1) nbtransi[i]=t(NREPLIC-1)*nbtransd[i]/sqrt(NREPLIC);
    else nbtransi[i]=0;
  }
  for(int i =0; i < NTTRANS; i++) {
    printf("\t* %28s : %10.2f\t+/- %10.2f\n", trans_name[i], nbtransm[i], nbtransi[i]);
  }
  printf("\t* %28s : %10.2f\t+/- %10.2f\n", "Total", nbtransm[NTTRANS], nbtransi[NTTRANS]);
  printf("\nMean response time\n\n");

  float mrt[NTTRANS+1], drt[NTTRANS+1], irt[NTTRANS+1];

  for (i=0; i<NTTRANS; i++) {
    mrt[i]=RTIMEC[i]/NREPLIC;
    drt[i]=(NREPLIC*RTIMEC2[i]-RTIMEC[i]*RTIMEC[i])/(NREPLIC*NREPLIC);
    if (drt[i]>0) drt[i]=sqrt(drt[i]);
    else drt[i]=0;
    if (NREPLIC>1) irt[i]=t(NREPLIC-1)*drt[i]/sqrt(NREPLIC);
    else irt[i]=0;    
  }

  if (nbtransm[NTTRANS]!=0) mrt[NTTRANS]=(transmgr->Mean(0)*transmgr->Mean(2))/nbtransm[NTTRANS];
  else mrt[NTTRANS]=0;
  if (nbtransd[NTTRANS]!=0) drt[NTTRANS]=(transmgr->Dev(0)*transmgr->Dev(2))/nbtransd[NTTRANS];
  else drt[NTTRANS]=0;
  if (NREPLIC>1) irt[NTTRANS]=t(NREPLIC-1)*drt[NTTRANS]/sqrt(NREPLIC);
  else irt[NTTRANS]=0;    
    
  
  for(int i =0; i < NTTRANS; i++) {
    printf("\t* %28s : %10.2f ms\t+/- %10.2f\n", trans_name[i], mrt[i], irt[i]);
  }
   printf("\t* %28s : %10.2f ms\t+/- %10.2f\n", "Total", mrt[NTTRANS], irt[NTTRANS]);


  printf("\nMean number of objects accessed\n\n");

  float nbobjm[NTTRANS+1], nbobjd[NTTRANS+1], nbobji[NTTRANS+1];

  for (i=0; i<NTTRANS+1; i++) {
    nbobjm[i]=NBOBJC[i]/NREPLIC;
    nbobjd[i]=(NREPLIC*NBOBJC2[i]-NBOBJC[i]*NBOBJC[i])/(NREPLIC*NREPLIC);
    if (nbobjd[i]>0) nbobjd[i]=sqrt(nbobjd[i]);
    else nbobjd[i]=0;
    if (NREPLIC>1) nbobji[i]=t(NREPLIC-1)*nbobjd[i]/sqrt(NREPLIC);
    else nbobji[i]=0;      
  }

  
  for(int i =0; i < NTTRANS; i++) {
    printf("\t* %28s : %10.2f\t+/- %10.2f\n", trans_name[i], nbobjm[i],nbobji[i]);
  }
   printf("\t* %28s : %10.2f\t+/- %10.2f\n", "Total", nbobjm[NTTRANS], nbobji[NTTRANS]);

  printf("\n*** CLUSTERING\n\n");

  float nbclm=NBCLC/NREPLIC;
  float nbcld=(NREPLIC*NBCLC2-NBCLC*NBCLC)/(NREPLIC*NREPLIC);
  if (nbcld>0) nbcld=sqrt(nbcld);
  else nbcld=0;
  float nbcli;
  if (NREPLIC>1) nbcli=t(NREPLIC-1)*nbcld/sqrt(NREPLIC);
  else nbcli=0;   
  //
  float ctimem=0;
  if (nbclm!=0) ctimem=CTIMEC/(nbclm*NREPLIC*1000);
  float ctimed=(NREPLIC*CTIMEC2-CTIMEC*CTIMEC)/(NREPLIC*NREPLIC*1000);
  if (ctimed>0) ctimed=sqrt(ctimed);
  else ctimed=0;
  if (nbcld!=0) ctimed=ctimed/nbcld;
  float ctimei;
  if (NREPLIC>1) ctimei=t(NREPLIC-1)*ctimed/sqrt(NREPLIC);
  else ctimei=0;  
  // ctimed mesure fausse !
  //
  float tctimem=TCTIMEC/(NREPLIC*1000);
  float tctimed=(NREPLIC*TCTIMEC2-TCTIMEC*TCTIMEC)/(NREPLIC*NREPLIC*1000);
  if (tctimed>0) tctimed=sqrt(tctimed);
  else tctimed=0;
  float tctimei;
  if (NREPLIC>1) tctimei=t(NREPLIC-1)*tctimed/sqrt(NREPLIC);
  else tctimei=0;  
  //
  float nccm=NCC/NREPLIC;
  float nccd=(NREPLIC*NCC2-NCC*NCC)/(NREPLIC*NREPLIC);
  if (nccd>0) nccd=sqrt(nccd); else nccd=0;
  float ncci;
  if (NREPLIC>1) ncci=t(NREPLIC-1)*nccd/sqrt(NREPLIC);
  else ncci=0;  
  //
  float tccm=TCC/NREPLIC;
  float tccd=(NREPLIC*TCC2-TCC*TCC)/(NREPLIC*NREPLIC);
  if (tccd>0) tccd=sqrt(tccd); else tccd=0;
  float tcci;
  if (NREPLIC>1) tcci=t(NREPLIC-1)*tccd/sqrt(NREPLIC);
  else tcci=0;  

  printf("\t* Mean number of reclusterings  : %10.2f  \t+/- %10.2f\n",nbclm,nbcli);
  printf("\t* Mean clustering duration      : %10.2f s\t+/- %10.2f\n",ctimem,ctimei);
  printf("\t* Mean cl.-rel. ops. overhead   : %10.2f s\t+/- %10.2f\n",tctimem,tctimei);
  printf("\t* Mean number of clusters built : %10.2f \t+/- %10.2f\n",nccm,ncci);
  printf("\t* Clusters' average size        : %10.2f o\t+/- %10.2f\n",tccm,tcci);

  //

  printf("\n*** THROUHPUT\n\n");

  float thrum=nbtransm[NTTRANS]*1000/simul->Tmax();
  //  float thrud=nbtransd[4]*1000/simul->Tmax(); // ?
  float thrui=nbtransi[NTTRANS]*1000/simul->Tmax(); // ?  

  printf("\t* Mean Throughput : %10.2f tr./s\t+/- %10.2f\n",thrum,thrui);

  //

  printf("\n*** I/Os\n\n");

  float nbiom=NBIOC/NREPLIC;
  float nbiod=(NREPLIC*NBIOC2-NBIOC*NBIOC)/(NREPLIC*NREPLIC);
  if (nbiod>0) nbiod=sqrt(nbiod);
  else nbiod=0;
  float nbioi;
  if (NREPLIC>1) nbioi=t(NREPLIC-1)*nbiod/sqrt(NREPLIC);
  else nbioi=0;    
  //
  float tnbiom=TNBIOC/NREPLIC;
  float tnbiod=(NREPLIC*TNBIOC2-TNBIOC*TNBIOC)/(NREPLIC*NREPLIC);
  if (tnbiod>0) tnbiod=sqrt(tnbiod);
  else tnbiod=0;
  float tnbioi;
  if (NREPLIC>1) tnbioi=t(NREPLIC-1)*tnbiod/sqrt(NREPLIC);
  else tnbioi=0;   
  //
  float cnbiom=CNBIOC/NREPLIC;
  float cnbiod=(NREPLIC*CNBIOC2-CNBIOC*CNBIOC)/(NREPLIC*NREPLIC);
  if (cnbiod>0) cnbiod=sqrt(cnbiod);
  else cnbiod=0;
  float cnbioi;
  if (NREPLIC>1) cnbioi=t(NREPLIC-1)*cnbiod/sqrt(NREPLIC);
  else cnbioi=0;   
  //
  float wnbiom=WNBIOC/NREPLIC;
  float wnbiod=(NREPLIC*WNBIOC2-WNBIOC*WNBIOC)/(NREPLIC*NREPLIC);
  if (wnbiod>0) wnbiod=sqrt(wnbiod);
  else wnbiod=0;
  float wnbioi;
  if (NREPLIC>1) wnbioi=t(NREPLIC-1)*wnbiod/sqrt(NREPLIC);
  else wnbioi=0;   
  //
  float pnbiom=PNBIOC/NREPLIC;
  float pnbiod=(NREPLIC*PNBIOC2-PNBIOC*PNBIOC)/(NREPLIC*NREPLIC);
  if (pnbiod>0) pnbiod=sqrt(pnbiod);
  else pnbiod=0;
  float pnbioi;
  if (NREPLIC>1) pnbioi=t(NREPLIC-1)*pnbiod/sqrt(NREPLIC);
  else pnbioi=0;   


  printf("\t* Mean number of I/Os      : %10.2f\t+/- %10.2f\n",
	 nbiom, nbioi);
  printf("\t* Read Transaction I/Os    : %10.2f\t+/- %10.2f\n",
	 tnbiom, tnbioi);
  printf("\t* Clustering I/Os          : %10.2f\t+/- %10.2f\n",
	   cnbiom, cnbioi);
  printf("\t* Write I/Os               : %10.2f\t+/- %10.2f\n",
	 wnbiom, wnbioi);
  printf("\t* Prefetch I/Os            : %10.2f\t+/- %10.2f\n",
	 pnbiom, pnbioi);
  

  printf("\n*** NUMBER OF PAGES\n\n");

  float nbpgm=NBPGC/NREPLIC;
  float nbpgd=(NREPLIC*NBPGC2-NBPGC*NBPGC)/(NREPLIC*NREPLIC);
  if (nbpgd>0) nbpgd=sqrt(nbpgd);
  else nbpgd=0;
  float nbpgi;
  if (NREPLIC>1) nbpgi=t(NREPLIC-1)*nbpgd/sqrt(NREPLIC);
  else nbpgi=0; 
  
  printf("\t* Mean number of pages used : %10.2f\t+/- %10.2f\n\n",nbpgm,nbpgi);

  //
  printf("\n*** Clustering Metrics\n\n");
  printf("\t* Working Set Size (win = %3d)  : %10.2f\n", WSS_WIN_SIZE,
	 (float) buffmgr->GetTotalDistinctPages() 
	 / (float) buffmgr->GetNumWinsItered());
  printf("\t* Long Term Expansion Factor    : %10.2f\n",
	 (float) buffmgr->GetLTEFNumDPages() / 
	 (float) _ltef_num_ideal_pages);


}

/////////////////////////////////////////////////////////////////////
// CLASSE User (RA1)
/////////////////////////////////////////////////////////////////////

// CLASSE User : Constructeur

User::User(char n[STRS], int cap, Simulation *sim, short id, Resource *rsp): Resource(n, cap, sim) {

  userid=id;
  rsproc=rsp;
  prevRoot = CRand(1,0,NO-1);
  currentTransactionRoot = prevRoot;
}

// CLASSE User : Evenement 0 - Arrivee sur RA1

void User::Event10(Client *client) {

  this->P(11,client,1);         // Reservation RA1
}

// CLASSE User : Evenement 1 - Temps de latence

void User::Event11(Client *client) {

  Sim()->Sched()->Schedule(12,Sim()->Tnow()+Exp(TRANS_THINK),client);
}

// CLASSE User : Evenement 2 - Attente processeur

void User::Event12(Client *client) {

  rsproc->P(13,client,1);
}

// CLASSE User : Evenement 3 - Generation transaction

void User::Event13(Client *client) {
  static int nnt =0;

  Client *newcli;

  newcli=GenTrans();            // R1
  rsproc->V();                  // Rendu processeur
  if (newcli->type <= NTTRANS)  // Transaction
    Sim()->Sched()->Schedule(20,Sim()->Tnow(),newcli);
  else                          // Groupement
    Sim()->Sched()->Schedule(30,Sim()->Tnow(),newcli);
  
#ifdef CONCURRENCY_ON
  Sim()->Sched()->Schedule(10,Sim()->Tnow(),client); // Nouvelle transaction
#endif;
  start_client = client;
  nnt++;
  this->V();                    // Rendu RA1
}

// CLASSE User : R1 - Generer transaction

Client *User::GenTrans() {
  
  Client *newcli;
  float p;

  // Nouveau client
  newcli=Sim()->NewClient();
  newcli->cliid=userid;


  if (MOVE_WINDTYPE == 0) {
    newcli->rootoid = DB->GetRandomOid(DIST5);
  }
  else if (MOVE_WINDTYPE == 1) {
    bool validRoot = false;
    while (validRoot == false) {
      newcli->rootoid = HMWD->NextRoot(prevRoot);
      if ( newcli->rootoid != NULL_OID ) {
	validRoot = true;
	prevRoot = newcli->rootoid;
      }
      else {
	prevRoot = DB->GetRandomOid(1);
      }
    }
  }
  else if (MOVE_WINDTYPE == 2) {
    bool validRoot = false;
    while (validRoot == false) {
      newcli->rootoid = HGMWD->NextRoot(prevRoot);
      if ( newcli->rootoid != NULL_OID ) {
	validRoot = true;
	prevRoot = newcli->rootoid;
      }
      else {
	prevRoot = DB->GetRandomOid(1);
      }
    }
  }
  else if (MOVE_WINDTYPE == 3) {
    bool validRoot = false;
    while (validRoot == false) {
      newcli->rootoid = HD->NextRoot(prevRoot);
      if ( newcli->rootoid != NULL_OID ) {
	validRoot = true;
	prevRoot = newcli->rootoid;
      }
      else {
	prevRoot = DB->GetRandomOid(1);
      }
    }
  }
  else if (MOVE_WINDTYPE == 4) {
    bool validRoot = false;
    while (validRoot == false) {
      newcli->rootoid = HHGD->NextRoot(prevRoot);
      if ( newcli->rootoid != NULL_OID ) {
	validRoot = true;
	prevRoot = newcli->rootoid;
      }
      else {
	prevRoot = DB->GetRandomOid(1);
      }
    }
  }
    if (MOVE_WINDTYPE == 5) {
    bool validRoot = false;
    while (validRoot == false) {
      newcli->rootoid = HCD->NextRoot(prevRoot);
      if ( newcli->rootoid != NULL_OID ) {
	validRoot = true;
	prevRoot = newcli->rootoid;
      }
      else {
	prevRoot = DB->GetRandomOid(1);
      }
    }
  }

  currentTransactionRoot = newcli->rootoid;

  if (HOTSELECTED == true) {
    newcli->hotTrans = true;
  }
  else {
    newcli->hotTrans = false;
  }
  root_vect.push_back(newcli->rootoid);

  newcli->trans_it=0;

  // Objet precedent (DSTC)
  newcli->prevoid=NULL_OID;

  // Choix du type de transaction
  p=Uni(0,1);
  newcli->reached_end = false;
  if (p<=PCT1) {                // Acces ensembliste
    newcli->type=1;
    newcli->curdepth=SETDEPTH;
  } else if (p<=PCT2) {         // Parcours simple
    newcli->type=2;
    newcli->curdepth=SIMDEPTH;
  } else if (p<=PCT3) {         // Parcours hierarchique
    newcli->type=3;
    newcli->reftype=CRand(1,0,NREFT-1);
    newcli->curdepth=HIEDEPTH;
  } else if (p<=PCC) {          // Parcours stochastique
    newcli->type=4;
    newcli->curdepth=STODEPTH;
  } else if (p<=PORAU) {
    newcli->type=5;
  } else if (p<=POI) {
    newcli->type=6;
  } else if (p<=POD) {
    newcli->type=7;
  } else if (p<=POSAU) {
    newcli->type=8;
  } else if (p<=PRI) {
    newcli->type=9;
  } else if (p<=PRD) {
    newcli->type=10;
  } else if (p <= PID) {
    //    cout << "selected seq oid" << endl;
    newcli->type=11;
    newcli->curdepth= SEQOIDDEPTH;
    newcli->rootoid = 0;
  }
  else newcli->type=GROUPT;        // Demande de groupement
  
  
  //  printf("Transaction (type #%d) created at time %f\n",newcli->type,Sim()->Tnow());

  return newcli;
}

/////////////////////////////////////////////////////////////////////
// CLASSE TransMgr (RA2)
/////////////////////////////////////////////////////////////////////

// CLASSE TransMgr : Constructeur

TransMgr::TransMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb, float t): Resource(n, cap, sim) {

  rsproc[0]=rsp;
  rsbd=rsb;
  tnet=t;
}


// returns the number of pages that the active portion of the DB maps to.
int TransMgr::GetNumIdealPages() {
  int total_size;

  total_size = 0;
  for (set<int>::const_iterator it = _ltef_distinct_objs.begin();
       it != _ltef_distinct_objs.end(); it++) {
    if (DB->GetObj(*it) != NULL) {
      total_size += DB->GetObj(*it)->ClassPtr()->GetSize();
    }
  }
  
  return (int)((float)total_size / (float) PGSIZE + 1.0);
}


// CLASSE TransMgr : Evenement 0 - Arrivee sur RA2

void TransMgr::Event20(Client *client) {
  this->P(21,client,1);         // Reservation RA2
}

// CLASSE TransMgr : Evenement 1 - Attente processeur (1)

void TransMgr::Event21(Client *client) {

  if (client->trans_it==0) {
    NBST[client->type-1]++;
    RTIME[client->type-1]-=Sim()->Tnow();
  }
  client->trans_it++;
  rsproc[client->cliid]->P(22,client,1);
}

// CLASSE TransMgr : Evenement 2 - Extraction objet / reservation BD

void TransMgr::Event22(Client *client) {

  ExtObj(client);               // R2
  
  if (MOVE_WINDTYPE == 1) {
    HMWD->InsertTraversedObject(currentTransactionRoot, client->rootoid);
  }
  else if (MOVE_WINDTYPE == 2) {
    HGMWD->InsertTraversedObject(currentTransactionRoot, client->rootoid);
  }
  else if (MOVE_WINDTYPE == 3) {
    HD->InsertTraversedObject(currentTransactionRoot, client->rootoid);
  }
  else if (MOVE_WINDTYPE == 4) {
    HHGD->InsertTraversedObject(currentTransactionRoot, client->rootoid);
  }
  else if (MOVE_WINDTYPE == 5) {
    HCD->InsertTraversedObject(currentTransactionRoot, client->rootoid);
  }

  rsproc[client->cliid]->V();   // Liberation processeur
  if (((client->type <= 4) || (client->type == 11)) && (client->curdepth<0)) {
    client->reached_end = true;
  }
  if (client->reached_end == false) {
      // Transaction en cours
    rsbd->P(23,client,1);       // Reservation BD
    _ltef_distinct_objs.insert(client->rootoid);
  }
  else {
    // Compte des trans. terminees
    NBTRANS[client->type-1]++;
    NBTRANS[NTTRANS]++;
    RTIME[client->type-1]+=Sim()->Tnow();
    this->V();                  // Rendu RA2
    Sim()->KillClient(client);  // Destruction client courant
    // if no currency is turned on then only start a new transaction when the
    // the previous one completes.  The code below starts a new transaction
    // when a transaction completes.
#ifndef CONCURRENCY_ON
  Sim()->Sched()->Schedule(10,Sim()->Tnow(),start_client); // Nouvelle transaction
#endif;
    
  }
}

// CLASSE TransMgr : Evenement 3 - Attente processeur (2)

void TransMgr::Event23(Client *client) {

  rsproc[client->cliid]->P(100,client,1);
}


void TransMgr::Event100(Client *client) {
  Sim()->Sched()->Schedule(24,Sim()->Tnow()+OBJ_THINK,client);
}

// CLASSE TransMgr : Evenement 4 - Acquisition des verrous

void TransMgr::Event24(Client *client) {

  Sim()->Sched()->Schedule(25,Sim()->Tnow()+GETLOCK,client);
}

// CLASSE TransMgr : Evenement 5 - Requete d'objet

void TransMgr::Event25(Client *client) {

  rsproc[client->cliid]->V();   // Liberation processeur
  // Requete d'objet
  Sim()->Sched()->Schedule(40,Sim()->Tnow(),client); // Requete transmise
}

// CLASSE TransMgr : Evenement 6 - Attente processeur (3)

void TransMgr::Event26(Client *client) {

  rsproc[client->cliid]->P(27,client,1);
}

// CLASSE TransMgr : Evenement 7 - Traitement transaction + rendu verrous

void TransMgr::Event27(Client *client) {

  TrTrans();                    // R6
  Sim()->Sched()->Schedule(28,Sim()->Tnow()+RELLOCK,client); // Rendu verrous
}

// CLASSE TransMgr : Evenement 8 - Transfert reseau

void TransMgr::Event28(Client *client) {

  rsproc[client->cliid]->V();   // Rendu processeur
  rsbd->V();                    // Rendu BD
  this->V();                   // Rendu RA2
  Sim()->Sched()->Schedule(30,Sim()->Tnow()+tnet*DB->GetObj(client->rootoid)->ClassPtr()->GetSize(),client); // Transfert reseau
}

// CLASSE TransMgr : R2 - Extraire objet

void TransMgr::ExtObj(Client *client) {
  

  switch(client->type) {
    
  case 1: NextSet(client);break;       // Acces ensembliste
  case 2: NextSimp(client);break;      // Parcours simple
  case 3: NextHier(client);break;      // Parcours hierarchique
  case 4: NextStoc(client);break;      // Parcours stochastique
  case 5: NextObjRAUpdate(client);break; // Random attribute update 
  case 6: NextObjInsert(client);break; // New object insertion
  case 7: NextObjDelete(client);break; // Object deletion
  case 8: NextObjSAUpdate(client);break; // Sequential attribute update 
  case 9: NextRefInsertion(client);break; // random reference insertion
  case 10: NextRefDeletion(client);break; // random reference deletion
  case 11: NextSeqOid(client);break;  // sequential access of objects via oid  
  default: client->curdepth=-1;        // Erreur => STOP
    
  }
  
  //  printf("Transaction type #%d - Depth %d, Next object: #%d (at time %f)\n",client->type,client->curdepth,client->rootoid,Sim()->Tnow());
}


// CLASSE TransMgr : R2.1 - Acces ensembliste

void TransMgr::NextSet(Client *client) {

  int i, j, k, n;

  if (client->curdepth==SETDEPTH) {     // Premier passage
    // Creation de la liste roots
    client->roots=new ChList();
    // Creation de la liste toaccess avec rootoid
    client->toaccess=new ChList();
    client->toaccess->Insert(1,client->rootoid);
    client->curdepth--;
  }
  // Traitement courant
  if (client->toaccess->NbElem()>0) {
    // Il reste des objets a acceder
    // On depile dans rootoid (prochain objet a acceder)
    client->rootoid=client->toaccess->Scan(1);
    client->toaccess->Delete(1);
    // L'objet accede devient une racine (derniere pos.)
    client->roots->Insert(client->roots->NbElem()+1,client->rootoid);
  } else {
    if (client->roots->NbElem()>0) { // Niveau suivant
      client->curdepth--;
      if (client->curdepth>=0) {
        // Tous les elements de roots sont depiles
        n=client->roots->NbElem();
        for (i=0; i<n; i++) {
          k=client->roots->Scan(1);
          client->roots->Delete(1);
          // Les objets pointes par cette racine sont ajoutes
          // dans toaccess
          for (j=0; j<DB->GetObj(k)->ClassPtr()->MaxNRef(); j++)
            if (DB->GetObj(k)->GetORef(j)!=NULL)
              client->toaccess->Insert(client->toaccess->NbElem()+1,DB->GetObj(k)->GetORef(j)->OID());
        }
        // Acces au premier objet de toaccess
        if (client->toaccess->NbElem()>0) {
          // On depile dans rootoid
          client->rootoid=client->toaccess->Scan(1);
          client->toaccess->Delete(1);
          // L'objet accede devient une racine
          client->roots->Insert(client->roots->NbElem()+1,client->rootoid);
        } else client->curdepth=-1; // Termine
      }
    } else client->curdepth=-1; // Termine
  }
  if (client->curdepth<0) { // Fin de transaction : liberation memoire
    delete client->roots;
    delete client->toaccess;
  }
}

// CLASSE TransMgr : R2.2 - Parcours simple

void TransMgr::NextSimp(Client *client) {

  int i;
  if (client->curdepth==SIMDEPTH) { // Premier passage
    // Creation liste toaccess en tant que pile avec rootoid au sommet
    client->toaccess=new ChList();
    client->toaccess->Insert(1,client->rootoid);
    // Creation liste roots => liste des profondeurs
    client->roots=new ChList();
    client->roots->Insert(1,SIMDEPTH-1);
  }
  // Traitement courant
  if (client->toaccess->NbElem()>0) {
    // L'objet a acceder est depile dans rootoid
    client->rootoid=client->toaccess->Scan(1);
    client->toaccess->Delete(1);
    // Sa profondeur dans curdepth
    client->curdepth=client->roots->Scan(1);
    client->roots->Delete(1);
    // Si la profondeur n'est pas atteinte, on empile les objets "enfants"
    if (client->curdepth>0) {
      for (i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; i>=0; i--)
        if (DB->GetObj(client->rootoid)->GetORef(i)!=NULL) {
	  
          client->toaccess->Insert(1, DB->GetObj(client->rootoid)->GetORef(i)->OID());
          client->roots->Insert(1,client->curdepth-1);
        }
    }
  } else { // Fin de parcours
    client->curdepth=-1;
    delete client->toaccess;
    delete client->roots;
  }
}

// CLASSE TransMgr : R2.3 - Parcours hierarchique

void TransMgr::NextHier(Client *client) {

  int i;

  if (client->curdepth==HIEDEPTH) { // Premier passage
    // Creation liste toaccess en tant que pile avec rootoid au sommet
    client->toaccess=new ChList();
    client->toaccess->Insert(1,client->rootoid);
    // Creation liste roots => liste des profondeurs
    client->roots=new ChList();
    client->roots->Insert(1,HIEDEPTH-1);
  }
  // Traitement courant
  if (client->toaccess->NbElem()>0) {
    // L'objet a acceder est depile dans rootoid
    client->rootoid=client->toaccess->Scan(1);
    client->toaccess->Delete(1);
    // Sa profondeur dans curdepth
    client->curdepth=client->roots->Scan(1);
    client->roots->Delete(1);
    // Si la profondeur n'est pas atteinte, on empile les objets "enfants"
    if (client->curdepth>0) {
      for (i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; i>=0; i--)
        if ((DB->GetObj(client->rootoid)->GetORef(i)!=NULL)
        && (DB->GetObj(client->rootoid)->ClassPtr()->GetTRef(i)==client->reftype)) {
          client->toaccess->Insert(1,DB->GetObj(client->rootoid)->GetORef(i)->OID());
          client->roots->Insert(1,client->curdepth-1);
        }
    }
  } else { // Fin de parcours
    client->curdepth=-1;
    delete client->toaccess;
    delete client->roots;
  }
}

// CLASSE TransMgr : R2.4 - Parcours stochastique

void TransMgr::NextStoc(Client *client) {

  int i;
  short k;
  float alea;

  client->curdepth--;
  if ((client->curdepth<(STODEPTH-1)) && (client->curdepth>=0)) { // Traitement courant
    // Tirage aleatoire
    // use uniform distribution
    alea=Uni(0,1);
    // Test de la relation a suivre (i)
    i=0;
    k=-1;
    while ((i<DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()) && (k<0)) {
      if (alea>STOCHTAB[i]) i++;
      else k=1;
    }
    if (k==1) {// Relation trouvee 
      if (DB->GetObj(client->rootoid)->GetORef(i)!=NULL) {
	// if an valid object was found for traversing.
        client->rootoid=DB->GetObj(client->rootoid)->GetORef(i)->OID();
      } else client->curdepth=-1; // Fin de parcours
    } else client->curdepth=-1;
  }
}


void TransMgr::NextObjRAUpdate(Client * client) {
  int temp_oid, temp_oid2;
  bool found_dup;
  
  if (client->toaccess == NULL) {
    client->toaccess = new ChList();
    client->toaccess->Insert(1,client->rootoid);
    for (int i = 2; i <= NUPDT; i++) {
      temp_oid=DB->GetRandomOid(DIST100);
      found_dup = false;
      for (int j = 1 ; j < i; j++) {
	temp_oid2 = client->toaccess->Scan(j);
	if (temp_oid == temp_oid2) {
	  found_dup = true;
	}
      }
      if (found_dup == true) {
	i--;
      }
      else {
	client->toaccess->Insert(i, temp_oid);
      }
    }
  }
  
  if (client->toaccess->NbElem()>0) {    
    client->rootoid=client->toaccess->Scan(1);
#ifdef SEE_OBJ_STATS
    cout << "obj size = " << 
      ((DB->GetObj(client->rootoid))->ClassPtr())->GetSize() << endl;
    int j =0;
    for (int i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; 
	 i>=0; i--) {
      if (DB->GetObj(client->rootoid)->GetORef(i) != NULL) {
	j++;
      }
    }
    cout << "obj number of ref. = " << j << endl;
#endif
    client->update_root_obj = true;
    client->toaccess->Delete(1);
  }
  else {
    delete client->toaccess;
    client->reached_end = true;
  }
}

// insert on new object
void TransMgr::NextObjInsert(Client * client) {

  if (client->update_root_obj == false) {
    client->rootoid = DB->InsertNewObj();
    DB->PlaceNewObj(client->rootoid);
    client->update_root_obj = true;
  }
  else {
    client->reached_end = true;
  }
}


// make an object unreachable
void TransMgr::NextObjDelete(Client * client) {
  // first mark the object as being updated (so it will be loaded up and 
  // marked as needing write I/O on it) then delete it.
  if (client->update_root_obj == false) {
    client->rootoid = DB->GetRandomOid(DIST102);
    client->update_root_obj = true;
  }
  else {
    // delete object from object graph
    DB->DeleteObj(client->rootoid);
    client->reached_end = true;
  }
}

void TransMgr::NextObjSAUpdate(Client * client) {
  int class_id;

  if (client->toaccess == NULL) {
    client->toaccess = new ChList();
    class_id = CRand(DIST103, 0, NC - 1 );
    for(int i = 1; i <= DB->GetClass(class_id)->Iterator()->NbElem(); i++) {
      client->toaccess->Insert(1, DB->GetClass(class_id)->Iterator()->Scan(i));
    }
  }
  
  if (client->toaccess->NbElem()>0) {    
    client->rootoid=client->toaccess->Scan(1);
#ifdef SEE_OBJ_STATS
    cout << "obj size = " << 
      ((DB->GetObj(client->rootoid))->ClassPtr())->GetSize() << endl;
    int j =0;
    for (int i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; 
	 i>=0; i--) {
      if (DB->GetObj(client->rootoid)->GetORef(i) != NULL) {
	j++;
      }
    }
    cout << "obj number of ref. = " << j << endl;
#endif
    client->update_root_obj = true;
    client->toaccess->Delete(1);
  }
  else {
    delete client->toaccess;
    client->reached_end = true;
  }
}



// WARNING:: when concurrent transactions are implemented, then must becareful
//           do the write locks at the correct places for this function.
void TransMgr::NextRefInsertion(Client * client) {
  int temp_oid;

  if (client->toaccess == NULL) {
    client->toaccess = new ChList();
    client->toaccess->Insert(1,client->rootoid);
    for (int i = 2; i <= NUPDT; i++) {
      temp_oid = DB->GetRandomOid(DIST100);
      client->toaccess->Insert(1,temp_oid);
    }
  }
  
  if (client->toaccess->NbElem()>0) {    
    client->rootoid=client->toaccess->Scan(1);
#ifdef SEE_OBJ_STATS
    cout << "obj size = " << 
      ((DB->GetObj(client->rootoid))->ClassPtr())->GetSize() << endl;
    int j =0;
    for (int i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; 
	 i>=0; i--) {
      if (DB->GetObj(client->rootoid)->GetORef(i) != NULL) {
	j++;
      }
    }
    cout << "obj number of ref. = " << j << endl;
#endif
    // note that insert reference may change the rootoid to a new object if
    // it determines that the maximum number of links has already been 
    // reached for the object. 
    DB->InsertRef(&client->rootoid, DIST100);
    client->update_root_obj = true;
    client->toaccess->Delete(1);
  }
  else {
    delete client->toaccess;
    client->reached_end = true;
  }
}



// WARNING:: when concurrent transactions are implemented, then must be careful
//           do the write locks at the correct places for this function.
void TransMgr::NextRefDeletion(Client * client) {
  int temp_oid;

  if (client->toaccess == NULL) {
    client->toaccess = new ChList();
    client->toaccess->Insert(1,client->rootoid);
    for (int i = 2; i <= NUPDT; i++) {
      temp_oid = DB->GetRandomOid(DIST100);
      client->toaccess->Insert(1,temp_oid);
    }
  }
  
  if (client->toaccess->NbElem()>0) {    
    client->rootoid=client->toaccess->Scan(1);
#ifdef SEE_OBJ_STATS
    cout << "obj size = " << 
      ((DB->GetObj(client->rootoid))->ClassPtr())->GetSize() << endl;
    int j =0;
    for (int i=DB->GetObj(client->rootoid)->ClassPtr()->MaxNRef()-1; 
	 i>=0; i--) {
      if (DB->GetObj(client->rootoid)->GetORef(i) != NULL) {
	j++;
      }
    }
    cout << "obj number of ref. = " << j << endl;
#endif
    // note that insert reference may change the rootoid to a new object if
    // it determines that the maximum number of links has already been 
    // reached for the object. 
    DB->DeleteRef(&client->rootoid, DIST100);
    client->update_root_obj = true;
    client->toaccess->Delete(1);
  }
  else {
    delete client->toaccess;
    client->reached_end = true;
  }
}


void TransMgr::NextSeqOid(Client * client) {

  if (client->curdepth == SEQOIDDEPTH) {
    client->toaccess = new ChList();
    client->toaccess->Insert(1, client->rootoid);

    for (int i = 1; i < SEQOIDDEPTH; i++) {  
      client->toaccess->Insert(1, i);
    }    
  }

  if (client->toaccess->NbElem() > 0) {
    client->rootoid = client->toaccess->Scan(1);
    client->toaccess->Delete(1);
  }
  else {
    client->curdepth = -1;
  }
  client->curdepth--;
}



// CLASSE TransMgr : R6 - Traiter transaction

void TransMgr::TrTrans() {

  // CODE A INSERER ICI pour traitement specifique
}

/////////////////////////////////////////////////////////////////////
// class DROClustObj
/////////////////////////////////////////////////////////////////////

DROClustObj::DROClustObj(int objId) {

  _objId = objId;
  _placementNext = NULL;
  _placementPrev = NULL;
  _sortNext = NULL;
}

DROClustObj * DROClustObj::getPlacementNext() {

  return _placementNext;

}

void DROClustObj::setPlacementNext(DROClustObj * placementNext) {

  _placementNext = placementNext;

}


DROClustObj * DROClustObj::getPlacementPrev() {

  return _placementPrev;

}

void DROClustObj::setPlacementPrev(DROClustObj * placementPrev) {

  _placementPrev = placementPrev;

}

int DROClustObj::getOid() {

  return _objId;

}

DROClustObj * DROClustObj::getSortNext() {

  return _sortNext;

}

void DROClustObj::setSortNext(DROClustObj * sortNext) {

  _sortNext = sortNext;

}

/////////////////////////////////////////////////////////////////////
// CLASSE ClustMgr (RA3)
/////////////////////////////////////////////////////////////////////

// CLASSE ClustMgr : Constructeur

ClustMgr::ClustMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb): Resource(n, cap, sim) {

  rsproc[0]=rsp;
  rsbd=rsb;
  Initialise();
  _clear_q_count = 0;
  _prp_rand_curr_page = 0;
}

void ClustMgr::Initialise() {
  freepage = NULL_PAGE_NUM;
  freepageleft = 0;
}

// CLASSE ClustMgr : Initialise le pointeur vers le BuffMgr

void ClustMgr::SetBuffPtr(BuffMgr *ptr) {

  buffmgr=ptr;
}

// CLASSE ClustMgr : Initialise le pointeur vers l'ObjMgr

void ClustMgr::SetObjPtr(ObjMgr *ptr) {

  objmgr=ptr;
}

// CLASSE ClustMgr : Evenement 0 - Arrivee sur RA3

void ClustMgr::Event30(Client *client) {

  if ((client->type>=1) && (client->type<=NTTRANS))
    this->P(31,client,1);       // Reservation RA3 pour traitement courant
  else if (client->type==GROUPT)
    this->P(34,client,1);       // Reservation RA3 pour groupement
}

// CLASSE ClustMgr : Evenement 1 - Tt : Attente processeur
// reserve the processor
void ClustMgr::Event31(Client *client) {

  rsproc[client->cliid]->P(32,client,1);
}



float ClustMgr::PRPRANDAnalysis(Client * client) {
  float time = 0.0;  
  bool foundinmem;
  int page_num;

  for (int i =0; i < PRPRANDRGRAIN; i++) {
    page_num = _prp_rand_curr_page;
    if (OPPORTUNISTIC_WRITE == 1) {
      foundinmem = buffmgr->NonEmptyUpdatedPageInMem(page_num);
      while (foundinmem == false) {
	page_num++;
	if (page_num < (DB->NbPages() -1)) {
	  foundinmem = buffmgr->NonEmptyUpdatedPageInMem(page_num);      
	}
	else {
	  page_num = 0;
	  foundinmem = buffmgr->NonEmptyUpdatedPageInMem(page_num);
	}
      }
    }   
    else if (OPPORTUNISTIC_READ == 1) {
      foundinmem = buffmgr->PageInMem(page_num);
      while ((foundinmem == false) && (DB->PageEmpty(page_num) == false)) {
	page_num++;
	if (page_num < (DB->NbPages() -1)) {
	  foundinmem = buffmgr->PageInMem(page_num);      
	}
	else {
	  page_num = 0;
	  foundinmem = buffmgr->PageInMem(page_num);
	}
      }
    }
    else {
      if (page_num >= (DB->NbPages() -2)) {
	page_num = 0;
      }
    }
    time += PRPRANDCPage(client, page_num);
  }
  return time;
}


// this needs to checked out before being used since it is a bit dodge at the
// moment

float ClustMgr::PRPRANDCPage(Client * client, int page_num) {
  float time = 0.0;
  bool foundinmem;
  bool foundinmem2;
  int selected_page;
  int temp_pg;
  multimap<const int,  part_info_t, int_cmp> * obj_seq;  
  obj_seq = new multimap<const int,  part_info_t , int_cmp>();
  set<int> reorg_scope_set;

  selected_page = page_num;
  if (OPPORTUNISTIC_WRITE == 1) {
    temp_pg = selected_page + 1;
    foundinmem = false;
    while ((foundinmem == false) && (temp_pg <= (DB->NbPages() - 1))
	   && (DB->PageEmpty(temp_pg) == false)) {
      foundinmem = buffmgr->NonEmptyUpdatedPageInMem(temp_pg);
      if (foundinmem == true) {
	reorg_scope_set.insert(selected_page);
	reorg_scope_set.insert(temp_pg);
	_prp_rand_curr_page = temp_pg;
      }
      temp_pg++;
      if (temp_pg == (DB->NbPages() - 1)) {
	selected_page = 0;
	foundinmem2 = buffmgr->NonEmptyUpdatedPageInMem(selected_page);
	while (foundinmem2 == false) {
	  selected_page++;	
	  foundinmem2 = buffmgr->NonEmptyUpdatedPageInMem(selected_page);      
	}
	temp_pg = selected_page + 1;
      }
    }    
  }
  else if (OPPORTUNISTIC_READ == 1) {  
    temp_pg = selected_page + 1;
    foundinmem = false;
    while ((foundinmem == false) && (temp_pg <= (DB->NbPages() - 1))
	   && (DB->PageEmpty(temp_pg) == true)) {
      foundinmem = buffmgr->PageInMem(temp_pg);
      if ((foundinmem == true) && (DB->PageEmpty(temp_pg) == true)) {
	reorg_scope_set.insert(selected_page);
	reorg_scope_set.insert(temp_pg);
	_prp_rand_curr_page = temp_pg;
      }
      temp_pg++;
      if (temp_pg == (DB->NbPages() - 1)) {
	selected_page = 0;
	foundinmem2 = buffmgr->PageInMem(selected_page);
	while ((foundinmem2 == false) && (DB->PageEmpty(selected_page) ==true)) {
	  selected_page++;	
	  foundinmem2 = buffmgr->PageInMem(selected_page);      
	}
	temp_pg = selected_page + 1;
      }
    }       
  }
  else {
    temp_pg = selected_page + 1;
    while  ((temp_pg <= (DB->NbPages() - 1))
	   && (DB->PageEmpty(temp_pg) == true)) {
      if ((DB->PageEmpty(temp_pg) == false)) {
	reorg_scope_set.insert(selected_page);
	reorg_scope_set.insert(temp_pg);
	_prp_rand_curr_page = temp_pg;
      }
      temp_pg++;
      if (temp_pg == (DB->NbPages() - 1)) {
	selected_page = 0;
	while (DB->PageEmpty(selected_page) == true) {
	  selected_page++;	
	}
	temp_pg = selected_page + 1;
      }
    }       
    /*   if (selected_page < (DB->NbPages() - 1)) {
      temp_pg = selected_page + 1; 
      reorg_scope_set.insert(selected_page);
      reorg_scope_set.insert(temp_pg);      
      _prp_rand_curr_page = temp_pg;
    }    
    else {
      selected_page=0;
      temp_pg = selected_page + 1; 
      reorg_scope_set.insert(selected_page);
      reorg_scope_set.insert(temp_pg);
      _prp_rand_curr_page = temp_pg;
      }*/
  } 

  PRP(&reorg_scope_set, obj_seq);  
  time += PlaceObj(&reorg_scope_set, obj_seq, client);
  delete obj_seq;

  return time;
}


float ClustMgr::PRPPRIOAnalysis(Client * client) {
  int page_num, top_page_num; 
  int * heat_array;
  float * waste_array;
  int * pn_array;
  float waste;
  int heat;
  float cb;
  float time = 0.0;
  set<int> * reorg_scope_set;
  
  
  heat_array = new int[PRPAGRAIN];
  waste_array = new float[PRPAGRAIN];
  pn_array = new int[PRPAGRAIN];

  for (int i =0 ; i < PRPAGRAIN; i++) {
    if (OPPORTUNISTIC_WRITE == 1)  {
      page_num = buffmgr->NextNonEmptyUpdatedPgNum(&STATS->GPBufferIter);
      if (page_num == NULL_PAGE_NUM) {
	buffmgr->StartIter(&STATS->GPBufferIter);
	page_num = buffmgr->NextNonEmptyUpdatedPgNum(&STATS->GPBufferIter);
	if (page_num == NULL_PAGE_NUM) {
	  cerr << "ERROR: page num should not be null, eventm.h" << endl;
	}
      }
    }   
    else if (OPPORTUNISTIC_READ == 1)  {
      page_num = buffmgr->NextNonEmptyPgNum(&STATS->GPBufferIter);
      if (page_num == NULL_PAGE_NUM) {
	buffmgr->StartIter(&STATS->GPBufferIter);
	page_num = buffmgr->NextNonEmptyPgNum(&STATS->GPBufferIter);
	if (page_num == NULL_PAGE_NUM) {
	  cerr << "ERROR: page num should not be null, eventm.h" << endl;
	}
      }
    }
    else {
      if (STATS->GPBufferIter < DB->NbPages()) {
	page_num = STATS->GPBufferIter;	
      }
      else {
	page_num =0;
	STATS->GPBufferIter = 0;
      }
      STATS->GPBufferIter++;
      if (STATS->GPBufferIter == freepage) {
	STATS->GPBufferIter++;
      }
    }  

    if (PQ->PageExists(page_num) == true) {
      PQ->RemovePage(page_num);
    }
    
    if (DB->GetPage(page_num) != NULL) {
      pn_array[i] = page_num;
      waste = DB->GetPage(page_num)->CalcWaste();
#ifdef DEBUG
      cout << "page = " << page_num << endl;
      cout << "waste = " << waste << endl;
#endif
      waste_array[i] = waste;
      heat = DB->GetPage(page_num)->CalcHeat();
      heat_array[i] = heat;
    }
  }
  
  
  for (int i =0 ; i < PRPAGRAIN; i++) {
    if (pn_array[i] != -1) {
     
      cb = (float) heat_array[i] * waste_array[i];
      // only if clustering badness (cb) is above the CB_THRESHOLD then
      // do we insert it into the priority queue
      if (cb > PRP_CB_THRESHOLD) {
	PQ->InsertElem(cb, pn_array[i]);
      }
    }
  }

  
  delete [] pn_array;
  delete [] waste_array;
  delete [] heat_array;
  
#ifdef DEBUG
  cout << "about to start taking from pq" << endl;
#endif
  for (int i =0; i < PRPPRIORGRAIN; i++) {
    top_page_num = PQ->PopTopElem();
    while ((top_page_num != NULL_PAGE_NUM) && 
	   (buffmgr->PageInMem(top_page_num) == false)) {      
      top_page_num = PQ->PopTopElem();
    }
    if (top_page_num == NULL_PAGE_NUM) {
      break;
    }
    reorg_scope_set = new set<int>;
    time += PRPPRIOCPage(client, top_page_num, reorg_scope_set);
    set<int>:: const_iterator iter;
    for (iter = reorg_scope_set->begin(); iter != reorg_scope_set->end();
	 iter++) {
      if ((*iter != top_page_num) && (PQ->PageExists(*iter) == true)) {
	if (PQ->RemovePage(*iter) == true) {
	  cb = (float) DB->GetPage(page_num)->CalcHeat() 
	    * DB->GetPage(page_num)->CalcWaste();
	    // only if clustering badness (cb) is above the CB_THRESHOLD then
	  // do we insert it into the priority queue
	  if (cb > PRP_CB_THRESHOLD) {
	    PQ->InsertElem(cb, *iter);
	  }
	}
	else {
	  cout << "ERROR: page about to removed can not be found" << endl;
	  exit(1);
	}
      }
    }
    delete reorg_scope_set;
  }
  _clear_q_count++;
  if ((_clear_q_count % PRP_CLEAR_QUEUE_NUM) == 0) {
    PQ->ClearQueue();
  }
  //    while ((top_page_num = PQ->PopTopElem()) != NULL_PAGE_NUM) {
  //  }
#ifdef DEBUG
  cout << "taken" << endl;
#endif
  return time;
}


float ClustMgr::PRPPRIOCPage(Client * client, int pg_num, 
			     set<int> * reorg_scope_set) {
  int temp_pg;
  ChList * objects;
  float time = 0.0;
  bool foundinmem;
  set<int>::iterator it;
  multimap<const int,  part_info_t, int_cmp> * obj_seq;
  bool page_empty;

  obj_seq = new multimap<const int,  part_info_t , int_cmp>();
  
  if (OPPORTUNISTIC_WRITE == 1) {
    if (pg_num > 0) {
      temp_pg = pg_num -1;
      foundinmem = false;
      while ((foundinmem == false) && (temp_pg >= 0)) {
	foundinmem = buffmgr->NonEmptyUpdatedPageInMem(temp_pg);
	if (foundinmem == true) {
	  reorg_scope_set->insert(temp_pg);
	}
	temp_pg--;
      }
    }
    
    // insert selected page in reorganisation scope, the insertion order
    // is important since all hottest objects need to be placed in lower
    // numbered pages
    reorg_scope_set->insert(pg_num); 

    if (pg_num < (DB->NbPages() - 1)) {
      temp_pg = pg_num +1;
      foundinmem = false;
      while ((foundinmem == false) && (temp_pg <= (DB->NbPages() - 1))) {
	foundinmem = buffmgr->NonEmptyUpdatedPageInMem(temp_pg);
	if (foundinmem == true) {
	  reorg_scope_set->insert(temp_pg);
	}
	temp_pg++;
      }
    }    
  }
  else if (OPPORTUNISTIC_READ == 1) {
    if (pg_num > 0) {
      temp_pg = pg_num -1;
      foundinmem = false;
      while ((foundinmem == false) && (temp_pg >= 0)) {
	foundinmem = buffmgr->NonEmptyPageInMem(temp_pg);
	if (foundinmem == true) {
	  reorg_scope_set->insert(temp_pg);
	}
	temp_pg--;
      }
    }
    
    // insert selected page in reorganisation scope, the insertion order
    // is important since all hottest objects need to be placed in lower
    // numbered pages
    reorg_scope_set->insert(pg_num); 

    if (pg_num < (DB->NbPages() - 1)) {
      temp_pg = pg_num +1;
      foundinmem = false;
      while ((foundinmem == false) && (temp_pg <= (DB->NbPages() - 1))) {
	foundinmem = buffmgr->NonEmptyPageInMem(temp_pg);
	if (foundinmem == true) {
	  reorg_scope_set->insert(temp_pg);
	}
	temp_pg++;
      }
    }    
  }
  else {
    if (pg_num > 0) {
      temp_pg = pg_num -1;
      page_empty = true;
      while ((page_empty == true) && (temp_pg >= 0)) {
	page_empty = DB->PageEmpty(temp_pg);
	if (page_empty == false) {
	  reorg_scope_set->insert(temp_pg);
	}
	temp_pg--;
      }
    }
    
    /*    if (pg_num > 0) {
      reorg_scope_set->insert(pg_num - 1);     
      } */

    // insert selected page in reorganisation scope, the insertion order is important
    // since all hottest objects need to be placed in lower numbered pages    
    reorg_scope_set->insert(pg_num); 
    
    
    if (pg_num < (DB->NbPages() - 1)) {
      temp_pg = pg_num +1;
      page_empty = true;
      while ((page_empty == true) && (temp_pg <= (DB->NbPages() - 1))) {
        page_empty = DB->PageEmpty(temp_pg);
	if (page_empty == false) {
	  reorg_scope_set->insert(temp_pg);
	}
	temp_pg++;
      }
    }    
    

    /*    if (pg_num < (DB->NbPages() - 1)) {
      reorg_scope_set->insert(pg_num + 1);      
      } */   
  }    
 

#ifdef DEBUG     
  cout << "Selected page: " << pg_num << endl;
  cout << "Reorg scope: ";
  for (it = reorg_scope_set->begin(); it != reorg_scope_set->end() ; it++) {
    cout << *it << " ";
  }
  cout << endl;

#endif
  
  for (it = reorg_scope_set->begin(); it != reorg_scope_set->end() ; it++) {  
    temp_pg = *it;
    objects = DB->GetPage(temp_pg)->Objects();
    for(int i = 1; i <= objects->NbElem(); i++) {
      client->rootoid=objects->Scan(i);
      // generate the I/O if neccessary
      time+=ObjRequest(client);
    }
  }

  PRP(reorg_scope_set, obj_seq);
  
  time += PlaceObj(reorg_scope_set, obj_seq, client);
  delete obj_seq;
  return time;
}

void ClustMgr::PRP(set<int> * reorg_scope_set,   
		   multimap<const int,  part_info_t, int_cmp> * obj_seq) {
  int page_num;
  part_info_t pi;
  ObjStats * os;
  int obj_id;
  int heat;
  ChList * objects;
  set<int> * temp_set;

  set<int>::iterator it;
  for (it = reorg_scope_set->begin(); it != reorg_scope_set->end() ; it++) {  
    page_num = *it;
    objects = DB->GetPage(page_num)->Objects();
    for(int i = 1; i <= objects->NbElem(); i++) {
      obj_id = objects->Scan(i);
      temp_set = new set<int>;
      temp_set->insert(obj_id);
      pi.obj_set = temp_set;
      pi.size = DB->GetObj(obj_id)->ClassPtr()->GetSize();
      if (pi.size > PGSIZE) {
	cerr <<"ERROR evemt.h: Object larger than a page, maybe because ";
	cerr <<"large objects are included" << endl;
      }
      os = STATS->GetObjStats(obj_id);
      if (os != NULL) {
	heat = os->GetHeat();
      }
      else {
	heat = 0;
      }
      obj_seq->insert(pair<const int, part_info_t>(heat, pi));
    }
  }

}


float ClustMgr::GPAnalysis(Client * client) {
  int page_num, top_page_num; 
  int * heat_array;
  int * et_array;
  int * pn_array;
  int et;
  int heat;
  float cb;
  float time = 0.0;
  set<int> * reorg_scope_set;

  heat_array = new int[GPAGRAIN];
  et_array = new int[GPAGRAIN];
  pn_array = new int[GPAGRAIN];

  for (int i =0; i < GPAGRAIN; i++) {
    heat_array[i] = -1;
    et_array[i] = -1;
    pn_array[i] = -1;
  }

  for (int i =0 ; i < GPAGRAIN; i++) {
    
    if (OPPORTUNISTIC_WRITE == 1) {      
      page_num = buffmgr->NextUpdatedPgNum(&STATS->GPBufferIter);
      if (page_num == NULL_PAGE_NUM) {
	buffmgr->StartIter(&STATS->GPBufferIter);
	page_num = buffmgr->NextUpdatedPgNum(&STATS->GPBufferIter);
	if (page_num == NULL_PAGE_NUM) {
	  cerr << "ERROR: page num should not be null, eventm.h" << endl;
	}
      }
    }
    else if (OPPORTUNISTIC_READ == 1) {
      page_num = buffmgr->NextPgNum(&STATS->GPBufferIter);
      if (page_num == NULL_PAGE_NUM) {
	buffmgr->StartIter(&STATS->GPBufferIter);
	page_num = buffmgr->NextPgNum(&STATS->GPBufferIter);
	if (page_num == NULL_PAGE_NUM) {
	  cerr << "ERROR: page num should not be null, eventm.h" << endl;
	}
      }
    } 
    else {
      if (STATS->GPBufferIter < DB->NbPages()) {
	page_num = STATS->GPBufferIter;	
      }
      else {
	page_num =0;
	STATS->GPBufferIter = 0;
      }
      STATS->GPBufferIter++;
      if (STATS->GPBufferIter == freepage) {
	STATS->GPBufferIter++;
      }
    }

    if (PQ->PageExists(page_num) == true) {
      PQ->RemovePage(page_num);
    }
 
    if (DB->GetPage(page_num) != NULL) {
      pn_array[i] = page_num;
      et = DB->GetPage(page_num)->CalcExtTension(buffmgr);
      et_array[i] = et;
      heat = DB->GetPage(page_num)->CalcHeat();
      heat_array[i] = heat;
#ifdef DEBUG
      cout << "page num = " << pn_array[i] << endl;	  
      cout << "et = " << et_array[i] << endl;
#endif
    }
  }
  
  
  for (int i =0 ; i < GPAGRAIN; i++) {
    if (pn_array[i] != -1) {
      //	cb = ((float) et_array[i] / (float) STATS->GetMaxPageExTension())
      //  * ((float) heat_array[i] / (float) STATS->GetMaxPageHeat()); 
      cb = (float) et_array[i] * (float) heat_array[i];
	// only if clustering badness (cb) is above the CB_THRESHOLD then
      // do we insert it into the priority queue
      if (cb > GP_CB_THRESHOLD) {
	PQ->InsertElem(cb, pn_array[i]);	
      }
    }
  }


  delete [] pn_array;
  delete [] et_array;
  delete [] heat_array;
  
#ifdef DEBUG
  cout << "about to start taking from pq" << endl;
#endif
  for (int i =0; i < GPRGRAIN; i++) {
    top_page_num = PQ->PopTopElem();
    while ((top_page_num != NULL_PAGE_NUM) && 
	   (buffmgr->PageInMem(top_page_num) == false)) {      
      top_page_num = PQ->PopTopElem();
    }
    if (top_page_num == NULL_PAGE_NUM) {
      break;
    }    
    reorg_scope_set = new set<int>;
    time += GPCPage(client, top_page_num, reorg_scope_set);  
    set<int>:: const_iterator iter;
    for (iter = reorg_scope_set->begin(); iter != reorg_scope_set->end();
	 iter++) {
      if ((*iter != top_page_num) && (PQ->PageExists(*iter) == true)) {
	if (PQ->RemovePage(*iter) == true) {
	  cb = (float) DB->GetPage(page_num)->CalcHeat() 
	    * (float)DB->GetPage(page_num)->CalcExtTension(buffmgr);
	      // only if clustering badness (cb) is above the CB_THRESHOLD then
	  // do we insert it into the priority queue
	  if (cb > GP_CB_THRESHOLD) {
	    PQ->InsertElem(cb, *iter);
	  }
	}
	else {
	      cout << "ERROR: page about to removed can not be found" << endl;
	      exit(1);
	}
      }
    }
    delete reorg_scope_set;
  }
  _clear_q_count++;
  if ((_clear_q_count % GP_CLEAR_QUEUE_NUM) == 0) {
    PQ->ClearQueue();
  }
  //  while ((top_page_num = PQ->PopTopElem()) != NULL_PAGE_NUM) {
  // }
  /*  et = DB->GetPage(top_page_num)->CalcExtTension(buffmgr);      
      heat = DB->GetPage(top_page_num)->CalcHeat();
      cb = ((float) et / (float) STATS->GetMaxPageExTension())
      * ((float) heat / (float) STATS->GetMaxPageHeat()); 
      cout << "cb  = " << cb << endl;
      }*/
#ifdef DEBUG
  cout << "taken" << endl;
#endif
  return time;
}


float ClustMgr::GPCPage(Client * client, int selected_page, 
			set<int> * reorg_scope_set) {
  float time = 0.0;
  
  int obj_id, ref_oid, iter, page_num;
  ObjStats * os;
  set<int> temp_set;
  // page numbers of reorganisation scope set
  set<int> temp_reorg_scope_set;
  ChList * objects;
  multimap<const int,  part_info_t, int_cmp> * obj_seq;

  obj_seq = new multimap<const int,  part_info_t , int_cmp>();

  objects = DB->GetPage(selected_page)->Objects();
  for(int i = 1; i <= objects->NbElem(); i++) {
    obj_id = objects->Scan(i);
    temp_set.insert(obj_id);
  }
#ifdef DEBUG
  cout << "about to recluster page: " << selected_page << endl;
#endif  
  for(int i = 1; i <= objects->NbElem(); i++) {
    obj_id = objects->Scan(i);
    os = STATS->GetObjStats(obj_id);
    if (os != NULL) {
      ref_oid = os->Start();
      while (ref_oid != NULL_OID) {
	if (temp_set.find(ref_oid) == temp_set.end()) {
	  temp_reorg_scope_set.insert(DB->GetObj(ref_oid)->PageDisk()->Scan(1));
	} 
	ref_oid = os->NextOid();
      }
    }
  }
#ifdef DEBUG
  cout << "temp_reorg_scope_set size = " << temp_reorg_scope_set.size() << endl;
#endif
  if (OPPORTUNISTIC_WRITE == 1) {
    buffmgr->StartIter(&iter);
    page_num = buffmgr->NextUpdatedPgNum(&iter);
    while (page_num != NULL_PAGE_NUM) {
#ifdef DEBUG
      cout << "page_num = " << page_num << endl;
#endif
      // if the current buffer page is one that is in a related page (page with       // tension crossing boundaries with the selected page) then 
      // it is in scope of reorganisation.
      if (temp_reorg_scope_set.find(page_num) != temp_reorg_scope_set.end()) {
	reorg_scope_set->insert(page_num);
      }
      //      cout << "hi" << endl;
      page_num = buffmgr->NextUpdatedPgNum(&iter);
    }
  } 
  else if (OPPORTUNISTIC_READ == 1) {
    buffmgr->StartIter(&iter);
    page_num = buffmgr->NextPgNum(&iter);
    while (page_num != NULL_PAGE_NUM) {
#ifdef DEBUG
      cout << "page_num = " << page_num << endl;
#endif
      // if the current buffer page is one that is in a related page (page with       // tension crossing boundaries with the selected page) then 
      // it is in scope of reorganisation.
      if (temp_reorg_scope_set.find(page_num) != temp_reorg_scope_set.end()) {
	reorg_scope_set->insert(page_num);
      }
      page_num = buffmgr->NextPgNum(&iter);
    }
  }
  else {
    *reorg_scope_set = temp_reorg_scope_set;
  }


  // now add the selected page into the scope of reorganisation
  reorg_scope_set->insert(selected_page);
  
  set<int>::iterator it;
#ifdef DEBUG
  cout << "Reorg scope: ";
  for (it = reorg_scope_set->begin(); it != reorg_scope_set->end() ; it++) {
    cout << *it << " ";
  }
  cout << endl;
#endif
  for (it = reorg_scope_set->begin(); it != reorg_scope_set->end() ; it++) {  
    page_num = *it;
    objects = DB->GetPage(page_num)->Objects();
    for(int i = 1; i <= objects->NbElem(); i++) {
      client->rootoid=objects->Scan(i);
      // generate the I/O if neccessary
      time+=ObjRequest(client);
      if (time < 0.0) {
	cout << "TIME IS NEG." << endl;
	cout << "oid = " << objects->Scan(i) << endl;
	exit(1);
      }
    }
  }

  // remove free page from list since the free page is just for placing
  // left over objects, that would not fit in pages the objects are initially
  // taken from.
  for (it = reorg_scope_set->begin(); it != reorg_scope_set->end() ; it++) { 
    if (page_num == freepage) {
      reorg_scope_set->erase(*it);
    }
  }

  // Choose the graph partitioning algorithm
  if (GPCALG == 0) {
    GreedyGP(reorg_scope_set, obj_seq);
  }
  time += PlaceObj(reorg_scope_set, obj_seq, client);
  delete obj_seq;
  return time;
}



float  ClustMgr::PlaceObj(set<int> * page_set, multimap<const int, part_info_t, int_cmp> * obj_seq, Client * client) {

  set<int>::const_iterator page_it;
  multimap<const int, part_info_t, int_cmp>::const_iterator obj_it;
  int size;
  vector<int> page_vect;
  vector<int> space_left_vect;
  Page * temp_page_ptr;
  Page * newpage;
  float time = 0.0;

#ifdef DEBUG
  ChList * objs; 
  int obj_id;
#endif

  // delete all objects from to be re-clustered pages
  // and move all page ids into a vector, since it is a better data structure
  // also initialise a vector for storing how much space is left in the page
  for (page_it = page_set->begin() ; page_it != page_set->end(); page_it++) {
    buffmgr->SetUpdated(*page_it);
#ifdef DEBUG
    objs = DB->GetPage(*page_it)->Objects();
    cout << "b page (" << *page_it << ") : ";
    for (int j = 1; j <= objs->NbElem(); j++) {
      cout << objs->Scan(j) << " ";
    }
    cout << endl;
#endif
    temp_page_ptr = DB->GetPage(*page_it);
    temp_page_ptr->DeleteAllObjs();
    temp_page_ptr->CreateNewObjList();
    page_vect.push_back(*page_it);
    space_left_vect.push_back(PGSIZE);
  
  }

  // says you do not know the free page is in memory
  bool free_page_in_mem = false; 
  set<int> * temp_obj_set;
  bool assigned;

  for (obj_it = obj_seq->begin(); obj_it != obj_seq->end(); obj_it++) {
    size = ((*obj_it).second).size;
    if (size > PGSIZE) {
      cerr << "ERROR: partition larger than a page" << endl;
      exit(1);
    }
    temp_obj_set = ((*obj_it).second).obj_set;
    // cout << "temp_obj_set = " << temp_obj_set << endl;
    assigned = false;
    for (unsigned int i =0; i < page_vect.size(); i++) {
      if ((space_left_vect[i] - size) > 0) {
	AssignPartToPage(temp_obj_set, page_vect[i]);
	space_left_vect[i] = space_left_vect[i] - size;
        delete temp_obj_set;
	assigned = true;
	break;
      }
    }
    // no existing page has enough room to fit the current partition
    // need to create a new page
    if (assigned == false) {
      if ((freepage == NULL_PAGE_NUM) || (freepageleft < size)) {
	if (freepage != NULL_PAGE_NUM) {
	  // this set updated is just for safety
	  buffmgr->SetUpdated(freepage);
	  page_vect.push_back(freepage);
	  space_left_vect.push_back(freepageleft);	
	}
	newpage = DB->NewPage();
	DB->AddPage(newpage);
	freepage = newpage->PageNum();
	
#ifdef PREFETCH_DEBUG
	cout << "making new free page" << endl;
#endif
	time+=AccDisk(freepage);	
	buffmgr->SetUpdated(freepage);
	AssignPartToPage(temp_obj_set, freepage);	
	freepageleft = PGSIZE - size;
	delete temp_obj_set;
      }
      else {
	buffmgr->SetUpdated(freepage);	
	AssignPartToPage(temp_obj_set, freepage);
	if (free_page_in_mem == false) {    
	  ChList * objects;
	  objects = DB->GetPage(freepage)->Objects();
	  client->rootoid=objects->Scan(1);
	  time+=ObjRequest(client);
	  free_page_in_mem = true;
	}
	freepageleft = freepageleft - size;
	delete temp_obj_set;
      }
    }
  }

  
  for (unsigned int i =0; i < page_vect.size(); i++) {
    DB->GetPage(page_vect[i])->ReCalcFree();
  }
  
  if (freepage != NULL_PAGE_NUM) {
    DB->GetPage(freepage)->ReCalcFree();
  }
  
#ifdef DEBUG  
  ObjStats * os;

  for (unsigned int i =0; i < page_vect.size(); i++) {
    objs = DB->GetPage(page_vect[i])->Objects();
    cout << "page (" << page_vect[i] << ") : ";
    for (int j = 1; j <= objs->NbElem(); j++) {
      obj_id = objs->Scan(j);
      cout << obj_id << " ";      
      cout << "(";   
      os = STATS->GetObjStats(obj_id);
      if (os != NULL) {
	 cout << os->GetHeat();
      }
      cout << ") (";
      Object * temp_obj_ptr2 = DB->GetObj(obj_id);
      cout << temp_obj_ptr2->PageDisk()->Scan(1) << ") ";
    }
    cout << endl;
  }
  
  if (freepage !=  NULL_PAGE_NUM) {
    objs = DB->GetPage(freepage)->Objects();
    cout << "free page (" << freepage << ") : ";

    for (int j = 1; j <= objs->NbElem(); j++) {      
      obj_id = objs->Scan(j);
      cout << obj_id << " ";      
      cout << "(";   
      os = STATS->GetObjStats(obj_id);
      if (os != NULL) {
	 cout << os->GetHeat();
      }
      cout << ") ";
    }
    cout << endl;
    
  }
#endif
  return time;
}

void ClustMgr::AssignPartToPage(set<int> * obj_set, int page_num) {
  int obj_id;
  set<int>::const_iterator it;
  Object * temp_obj_ptr;
  Page * temp_page_ptr;

  temp_page_ptr = DB->GetPage(page_num);
  for (it = obj_set->begin(); it != obj_set->end(); it++) {
    obj_id = *it;
    temp_page_ptr->InsertObj(obj_id);
    temp_obj_ptr = DB->GetObj(obj_id);
    temp_obj_ptr->DeleteAllPageDisks();
    temp_obj_ptr->CreateNewPageDisk();
    temp_obj_ptr->InsertPage(page_num);
  }
}

bool EnoughSpace(set<int> * set1, set<int> * set2);

// Algorithm from: Partition-Based Clustering in Object Base From Theory to
// practice,  1994
void ClustMgr::GreedyGP(set<int> * page_set, 
			multimap<const int, part_info_t, int_cmp> * obj_seq) {
  int page_num;
  ChList * objects;
  set<int> * part;
  hash_map<int, set<int> *, hash<int>, eqoid> part_map;   
  ObjStats * os;
  int ref_oid;
  set<int> reorg_set;
  pair_t temp_pair;
  int obj_id;
  int space;

  multimap<int, pair_t, int_cmp> tension_map;

  set<int>::iterator it;
  for (it = page_set->begin(); it != page_set->end() ; it++) {  
    page_num = *it;
    objects = DB->GetPage(page_num)->Objects();
    for(int i = 1; i <= objects->NbElem(); i++) {
      obj_id = objects->Scan(i);
      part = new set<int>;
      part->insert(obj_id);
      part_map[obj_id] = part;
      reorg_set.insert(obj_id);
    }
  }
  
  set<int>::const_iterator oiter;
  for (oiter = reorg_set.begin(); oiter != reorg_set.end();
       oiter++) {
    obj_id = *oiter;
    os = STATS->GetObjStats(obj_id);
    if (os != NULL) {
      ref_oid = os->Start();
      while (ref_oid != NULL_OID) {
	// only include tension pairs of objects in the scope of 
        // re-organisation
	if (reorg_set.find(ref_oid) != reorg_set.end()) {
	  temp_pair.from_oid = obj_id;
	  temp_pair.to_oid = ref_oid;
	  tension_map.insert(pair<const int, pair_t>(os->GetCurrTension(),
						     temp_pair));
	}
	ref_oid = os->NextOid();
      }
    }
  }

  multimap<const int, pair_t, int_cmp>::const_iterator ten_iter;
  set<int> * from_set;
  set<int> * to_set;
  set<int> * result_set;

  for(ten_iter = tension_map.begin(); ten_iter != tension_map.end();
      ten_iter++) {
    temp_pair = ten_iter->second;

    if ((part_map.find(temp_pair.from_oid) != part_map.end())  && 
	(part_map.find(temp_pair.to_oid) != part_map.end())) {
      from_set = part_map[temp_pair.from_oid];
      to_set = part_map[temp_pair.to_oid];

      if ((to_set != from_set) && (EnoughSpace(from_set, to_set) == true)) {
	result_set = new set<int>();
	set_union(from_set->begin(), from_set->end(), to_set->begin(), 
		  to_set->end(), inserter(*result_set, result_set->begin()));
	delete from_set;
	delete to_set;
	
	for  (set<int>::iterator it = result_set->begin(); it != result_set->end(); it++) {
	  part_map[*it] = result_set;
	}	
      }
    }
  }
  
  set<int> * temp_set;
  hash_map<int, set<int> *, hash<int>, eqoid>::const_iterator part_iter;
  part_info_t pi;
  hash_set<int, hash<int>, eqoid> partProcessed;
  for (part_iter = part_map.begin(); part_iter != part_map.end();
       part_iter++) {
    temp_set = part_iter->second;

    if (partProcessed.find((int)temp_set) == partProcessed.end()) {
      partProcessed.insert((int)temp_set);
      set<int>::iterator it1;
      
      space = 0;
      
      int heat = 0;
      
      for (it1 = temp_set->begin(); it1 != temp_set->end(); it1++) {
	space +=DB->GetObj(*it1)->ClassPtr()->GetSize();
	os = STATS->GetObjStats(*it1);
	if (os != NULL) {
	  heat += os->GetHeat();
	}
      }
      pi.size = space;
      if (space > PGSIZE) {
	cerr << "ERROR: Partition larger than a page, maybe because large objects are included" << endl;
      }
      pi.obj_set = temp_set;
      if (GP_TYPE == 0) {
	obj_seq->insert(pair<const int, part_info_t>(space, pi));
      }
      else if (GP_TYPE == 1) {
	obj_seq->insert(pair<const int, part_info_t>(heat, pi));
      }
#ifdef DEBUG
      cout << "set: ";
      for (it1 = temp_set->begin(); it1 != temp_set->end(); it1++) {
	cout << *it1 << " ";
      }
      cout << endl;
#endif
    }
  }
}

// precondition: all object sizes are less than a page 
bool EnoughSpace(set<int> * set1, set<int> * set2) {
  set<int>::iterator iter;
  bool enough_space = true;
  int sp =0;    

  for (iter = set1->begin(); iter != set1->end(); iter++) {
    sp +=DB->GetObj(*iter)->ClassPtr()->GetSize();
    //    cout << "obj 1 = " << *iter << endl;
  }

  for (iter = set2->begin(); iter != set2->end(); iter++) {
    sp +=DB->GetObj(*iter)->ClassPtr()->GetSize();
    //  cout << "obj 2 = " << *iter << endl;
    // cout << "sp = " << sp << endl;
    if (sp > PGSIZE) {
      enough_space = false;
      break;
    }
  }
  // cout <<"out" << endl;
  
  return enough_space;
}

// CLASSE ClustMgr : Evenement 2 - Tt : R7

void ClustMgr::Event32(Client *client) {

  float cltt_time = 0;

  if (CLUSTP == 9) {
    cltt_time=DSTCCollect(client);              // R7
    TCTIME+=cltt_time;
  }
  else if (CLUSTP == 1) {
    cltt_time=GPCollect(client);
    TCTIME+=cltt_time;
  }
  else if (CLUSTP == 2) {
    cltt_time=PRPCollect(client);
    TCTIME+=cltt_time;    
  }
  else if (CLUSTP == 3) {
    cltt_time=DROCollect(client);
    TCTIME+=cltt_time;        
  }
#ifdef DEBUG
  cout << "cltt_time = " << cltt_time << endl;
#endif
  Sim()->Sched()->Schedule(33,Sim()->Tnow()+cltt_time,client); // Fin traitement
}

float ClustMgr::PRPCollect(Client *client) {
  float time = 0.0;
  Client * newcli;

  time=0.0;
  DB->ni++;
  STATS->IncStat(NULL_OID, client->rootoid, heat, 1);
  
  if (DB->ni>= PRPNUMACCESS) {
    // printf("DSTC: Selection + Consolidation at time %f\n",Sim()->Tnow());
    // Creation client demande de regroupement
    // create a new clustering client (process)
    newcli=Sim()->NewClient();
    newcli->cliid=client->cliid;
    newcli->type=GROUPT;
    DB->ni = 0;
    Sim()->Sched()->Schedule(30,Sim()->Tnow(),newcli);
  }
  
  return time;
}


float ClustMgr::GPCollect(Client *client) {
  float time;
  Client * newcli;

  time=0.0;
  DB->ni++;
  STATS->IncStat(client->prevoid, client->rootoid, trans, 1);
  client->prevoid = client->rootoid;
  if (DB->ni>= GPNUMACCESS) {
    // printf("DSTC: Selection + Consolidation at time %f\n",Sim()->Tnow());
    // Creation client demande de regroupement
    // create a new clustering client (process)
    newcli=Sim()->NewClient();
    newcli->cliid=client->cliid;
    newcli->type=GROUPT;
    DB->ni = 0;
    Sim()->Sched()->Schedule(30,Sim()->Tnow(),newcli);
  }
  
  return time;
}



// CLASSE ClustMgr : Evenement 3 - Tt : Fin de traitement

void ClustMgr::Event33(Client *client) {

  rsproc[client->cliid]->V();   // Liberation processeur
  this->V();                    // Liberation gest. de groupement (RA3)
  Sim()->Sched()->Schedule(20,Sim()->Tnow(),client); // Suite de la transaction
}


// CLASSE ClustMgr : Evenement 4 - Gr : Blocage BD
// I think BD is db (database) in english
void ClustMgr::Event34(Client *client) {

  Client *ficli;
  int i;

  // Creation de MULTILVL-1 clients fictifs pour bloquer la BD
  for (i=0; i<MULTILVL-1; i++) {
    ficli=Sim()->NewClient();
    ficli->type=VCT;
    rsbd->P(35,ficli,9);
  }
  // Achevement du blocage
  rsbd->P(36,client,8);
}

// CLASSE ClustMgr : Evenement 5 - Gr : Destruction clients fictifs

void ClustMgr::Event35(Client *client) {

  Sim()->KillClient(client);
}

// CLASSE ClustMgr : Evenement 6 - Gr : Attente processeur

void ClustMgr::Event36(Client *client) {

  rsproc[client->cliid]->P(37,client,1);
}

// CLASSE ClustMgr : Evenement 7 - Gr : R8

void ClustMgr::Event37(Client *client) {

  float clustering_time;

  clustering_time=Cluster(client);              // R8
  // printf("Clustering time=%f (end at time %f)\n",clustering_time,Sim()->Tnow()+clustering_time);
  CTIME+=clustering_time;
#ifdef DEBUG
  cout << "clustering_time = " << clustering_time << endl;
#endif
  Sim()->Sched()->Schedule(38,Sim()->Tnow()+clustering_time,client); // Fin de clustering
}

// CLASSE ClustMgr : Evenement 8 - Gr : Fin clustering

void ClustMgr::Event38(Client *client) {

  int i;

  rsproc[client->cliid]->V();   // Rendu processeur
  for (i=0; i<MULTILVL; i++) rsbd->V(); // Liberation BD entiere
  this->V();                    // Liberation RA3
  Sim()->KillClient(client);    // OUT
  if (CLUSTP==VCT) {
    NBCL++;
  }
}

float ClustMgr::DROCollect(Client * client) {
  float time;
  int currOid;
  int currPid;
  Client * newcli;


  time = 0.0;
  currOid = client->rootoid;
  DB->ni++;

  // only consider small objects
  currPid = DB->GetObj(currOid)->PageDisk()->Scan(1);
  
  if (DROPageStatsMap[currPid]->setObjUsageIndicator(currOid, true) == false) {
    
    cerr << " Object does not exist when setting usage indicator in eventm.h";
    cerr << endl;
    exit(1);

  }

  if (DROPageStatsMap[currPid]->incObjAccessFreq(currOid) == false) {
    
    cerr << " Object does not exist when incrementing freq in eventm.h";
    cerr << endl;
    exit(1);

  }

  if (DB->ni>= DRONUMACCESS) {
    
    // printf("DSTC: Selection + Consolidation at time %f\n",Sim()->Tnow());
    // Creation client demande de regroupement
    // create a new clustering client (process)
    newcli=Sim()->NewClient();
    newcli->cliid=client->cliid;
    newcli->type=GROUPT;
    DB->ni = 0;
    Sim()->Sched()->Schedule(30,Sim()->Tnow(),newcli);
  }
  
  
  return time;
}

// CLASSE ClustMgr : R7 - Effectuer traitement relatif au groupement, retourne le temps necessaire
//                        Strategie DSTC
//                        (Phases d'Observation, de Selection et de
//                        Consolidation)

// The observation phase works by updating 

float ClustMgr::DSTCCollect(Client *client) {

  int place, i, j, k, i1, i2, ifc, gflag;
  short stop;
  Client *newcli;
  float fe, time;

  // Phase d'Observation

  time=0;

  //  cout <<"HHHHHHHHHHHH" << endl;
  // Mise a jour de VO
  DB->vo[client->rootoid]++;

  // S'il existe un objet precedemment accede, MaJ de MO
  if ((client->prevoid != NULL_OID) && (client->prevoid != client->rootoid)) {
    // Nouvelle entree
    DB->ni++;
    // Recherche si le compteur existe deja
    place=1;
    stop=0;  
    
    // The entries in mo[][0] are stoed in order of size starting with the
    // smallest one first 
    while ((place<=DB->mo[client->prevoid][0]->NbElem()) && (!stop)) {
      /*cout << "mo[client->prevoid][0]->Scan(place) = " << 
	DB->mo[client->prevoid][0]->Scan(place) << endl;
	cout << "client->rootoid = " << client->rootoid << endl;*/
      if (DB->mo[client->prevoid][0]->Scan(place)<client->rootoid) place++;
      else stop=1;
    }
    if (stop) {
      if (DB->mo[client->prevoid][0]->Scan(place)==client->rootoid) {
	DB->mo[client->prevoid][1]->Update(place,DB->mo[client->prevoid][1]->Scan(place)+1);
      } else { // Compteur a inserer en place (val=1)
	DB->mo[client->prevoid][0]->Insert(place,client->rootoid);
	DB->mo[client->prevoid][1]->Insert(place,1);
      }
    } else { // Compteur a inserer en queue (place)
	DB->mo[client->prevoid][0]->Insert(place,client->rootoid);
	DB->mo[client->prevoid][1]->Insert(place,1);
    }
  }
  // Evolution objet precedent
  client->prevoid=client->rootoid;
  // printf("DSTC: Observation entry #%d at time %f\n",DB->ni,Sim()->Tnow());

  // Phases de Selection et de Consolidation
  
  if (DB->ni>=N) {
    // printf("DSTC: Selection + Consolidation at time %f\n",Sim()->Tnow());
    // Creation client demande de regroupement
    // create a new clustering client (process)
    newcli=Sim()->NewClient();
    newcli->cliid=client->cliid;
    newcli->type=GROUPT;
    newcli->o1=new ChList;
    newcli->o2=new ChList;
    newcli->fc=new ChList;
    //  Parcours de MO
    for (i=0; i<NO; i++) {
      for (k=1; k<=DB->mo[i][0]->NbElem(); k++) {
        j=DB->mo[i][0]->Scan(k);
        if ((i!=j) && ((DB->vo[i]>=TFA) || (DB->vo[j]>=TFA))) { // Mesures valides
          // Calcul de MO(i,j)/VO(i)
          if (DB->vo[i]>0) fe=1.0*DB->mo[i][1]->Scan(k)/DB->vo[i];
          else fe=0;
          // Recherche de MO(j,i)
          place=1;
          stop=0;
          while ((place<=DB->mo[j][0]->NbElem()) && (!stop))
            if (DB->mo[j][0]->Scan(place)<i) place++;
            else stop=1;
          if (stop)
            if (DB->mo[j][0]->Scan(place)==i) {
              // Calcul de MO(j,i)/VO(j)
              if (DB->vo[j]>0) fe+=1.0*DB->mo[j][1]->Scan(place)/DB->vo[j];
              // Suppression MO(j,i) pour ne pas refaire le calcul
              DB->mo[j][0]->Delete(place);
              DB->mo[j][1]->Delete(place);
            }
          // printf("fe(%d,%d)=%f\n",i,j,fe);
          if (fe>=TFE) { // Mesures valides pour la matrice elementaire ME
            // MC est triangulaire => ordonnancement des indices
            if (i<j) {
              i1=i;
              i2=j;
            } else {
              i1=j;
              i2=i;
            }
            // Recherche de l'ancien fc dans MC
            place=1;
            stop=0;
            while ((place<=DB->mc[i1][0]->NbElem()) && (!stop))
              if (DB->mc[i1][0]->Scan(place)<i2) place++;
              else stop=1;
            if (stop) {
              if (DB->mc[i1][0]->Scan(place)==i2) {
                // Ancien fc trouve => maj
                ifc=(int)(W*DB->mc[i1][1]->Scan(place)+(1-W)*fe*100);
                DB->mc[i1][1]->Update(place,ifc);
                DB->mc[i1][2]->Update(place,DB->pi);
                gflag=DB->mc[i1][3]->Scan(place);
              } else {
                // Nouveau facteur fc
                ifc=(int)(fe*100);
                gflag=0;
                DB->mc[i1][0]->Insert(place,i2);
                DB->mc[i1][1]->Insert(place,ifc);
                DB->mc[i1][2]->Insert(place,DB->pi);
                DB->mc[i1][3]->Insert(place,gflag);
              }
            } else {
              // Nouveau facteur fc
              ifc=(int)(fe*100);
              gflag=0;
	      //cout << "place = " << place << endl;
	      //cout << "DB->mc[i1][0]->NbElem() = " << DB->mc[i1][0]->NbElem()
	      //	   << endl;
              DB->mc[i1][0]->Insert(place,i2);
	      //cout << "DB->mc[i1][1]->NbElem() = " << DB->mc[i1][1]->NbElem()
	      //<< endl;
              DB->mc[i1][1]->Insert(place,ifc);
	    
              DB->mc[i1][2]->Insert(place,DB->pi);
              DB->mc[i1][3]->Insert(place,gflag);
            }
           //printf("ifc=%d, TFC*100=%d, gflag=%d\n",ifc,(int)(TFC*100),gflag);
            if ((ifc>=TFC*100) && (!gflag)) {
              // Mise a jour du flag de groupement
              place=1;
              stop=0;
	      /*while ((place<=DB->mc[i1][0]->NbElem()) && (!stop))
                if (DB->mc[i1][0]->Scan(place)<i2) place++;
                else stop=1;
              if (stop)
                if (DB->mc[i1][0]->Scan(place)==i2)
		DB->mc[i1][3]->Update(place,1);*/
              // Ajout du couple (i1, i2) dans la liste des objets
              // a regrouper	      
	      // First check if the objects in question has already been 
	      // deleted or not.
	      if ((DB->GetObj(i1) != NULL) && (DB->GetObj(i2) != NULL)) {
		if (((OPPORTUNISTIC_READ == 0) 
		     || ((buffmgr->ObjInMem(i1) == true)
			 && (buffmgr->ObjInMem(i2) == true)))
		    && ((OPPORTUNISTIC_WRITE == 0) ||
			((buffmgr->ObjInUpdatedMem(i1) == true)
			 && (buffmgr->ObjInUpdatedMem(i2) == true)))){
		  newcli->o1->Insert(1,i1);
		  newcli->o2->Insert(1,i2);
		  newcli->fc->Insert(1,ifc);
		  //		  cout << "clust ( " << i1 << "," << i2 << " ), ";
		}
	      }
            }
          }
        }
      }
    }
    // cout << endl;
    // DB->DisplayMO();
    // DB->DisplayMC();
    DB->NewPeriod();
    // Declenchement du regroupement
    if (newcli->o1->NbElem()>0) {
      /* cout << "newcli->o1->NbElem() = " << newcli->o1->NbElem() << endl;
      cout <<"*********o1**********" << endl;
      newcli->o1->Display();
      cout <<"*********o2**********" << endl;
      newcli->o2->Display();
      cout <<"*********tf**********"<< endl;
      newcli->fc->Display();*/
      Sim()->Sched()->Schedule(30,Sim()->Tnow(),newcli);
    }
    else {
      delete newcli->o1;
      delete newcli->o2;
      delete newcli->fc;
      Sim()->KillClient(newcli);
    }
  }

  return time;
}

// CLASSE ClustMgr : R8 - Effectuer groupement, retourne le temps necessaire

float ClustMgr::Cluster(Client *client) {

  lastpage=-999;
  if (CLUSTP==9) { // Politique de groupement definie par l'utilisateur
    printf("DSTC clustering triggered at time %f\n",Sim()->Tnow());
    return DSTCCluster(client);
  }
  else if (CLUSTP==1) {
    //    cout << "GP clustering triggered " << endl;
    return GPCluster(client);
  } else if (CLUSTP==2)  { 
    //   cout << "PRP clustering triggered " << endl;
    return PRPCluster(client);
  }
  else if (CLUSTP == 3) {
    //  cout << "DRO clustering triggered " << endl;
    return DROCluster(client);
  }

  else return 0.0; // Sinon, pas de traitement
}


// CLASSE ClustMgr : Groupement par l'algorithme DSTC, retourne le temps necessaire

float ClustMgr::DSTCCluster(Client *client) {

  ChList *ur[3];                        // Unite de regroupement (UR)  // clustering unit 
                                        // 0 - OID objet 1
                                        // 1 - OID objet 2
                                        // 2 - Poids (fc)
  ChList *cluster;                      // Cluster courant
  int i, j, k, n, place, place2, rsize, v1, v2, v3, v4, obj;
  short stop, stop2;
  Page *newpage, *cpage;
  float time;

  time=0;

  //  cout << "DSTCDSTC" << endl;
  for (int i =0 ; i < 3; i++) {
    ur[i] = NULL;
  }
  //printf("DSTC! %d\n",client->o1->NbElem());

  // Enrichissement de la liste (o1,o2) a partir de MC

  n=client->o1->NbElem();
  place=n+1;
  // Parcours de la liste (o1,o2) actuelle
  for (i=1; i<=n; i++) {
    // o1 
    v1=client->o1->Scan(i);
    

    // Parcours de MC pour l'objet courant v1
    for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) {	
      
      v2=DB->mc[v1][0]->Scan(j);
      if (DB->mc[v1][1]->Scan(j)>=TFC) {
        // Parcours de (o1,o2) pour voir si le couple (v1,v2)
        // ne s'y trouve pas deja
        k=1;
        stop=0;
        while ((ur[0]!=NULL) &&(k<=ur[0]->NbElem()) && (!stop)) {
	  
          v3=client->o1->Scan(k);
          v4=client->o2->Scan(k);
          if ( ((v1==v3) && (v2==v4)) || ((v1==v4) && (v2==v3)) ) stop=1;
          else k++;
        }	
	

        //printf("FIN while ((k<=ur[0]->NbElem()) && (!stop)) j=%d\n",j);
	
	// First check if the objects in question has already been 
	// deleted or not.
	if ((DB->GetObj(v1) != NULL) && (DB->GetObj(v2) != NULL)) {
	  if ((!stop) &&(((OPPORTUNISTIC_READ == 0) 
			  || ((buffmgr->ObjInMem(v1) == true)
			      && (buffmgr->ObjInMem(v2) == true)))
			 && ((OPPORTUNISTIC_WRITE == 0) ||
			     ((buffmgr->ObjInUpdatedMem(v1) == true)
			      && (buffmgr->ObjInUpdatedMem(v2) == true))))) {
	    client->o1->Insert(place,v1);
	    client->o2->Insert(place,v2);
	    client->fc->Insert(place,DB->mc[v1][1]->Scan(j));
	    place++;
	  }
	}
      }
    }
    
    
    //printf("FIN for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) i=%d\n",i);
    // o2
    v1=client->o2->Scan(i);
    // Parcours de MC pour l'objet courant v1
    for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) {
      v2=DB->mc[v1][0]->Scan(j);
      if (DB->mc[v1][1]->Scan(j)>=TFC) {
        // Parcours de (o1,o2) pour voir si le couple (v1,v2)
        // ne s'y trouve pas deja
        k=1;
        stop=0;
        while ((ur[0] != NULL) && (k<=ur[0]->NbElem()) && (!stop)) {
          v3=client->o1->Scan(k);
          v4=client->o2->Scan(k);
          if ( ((v1==v3) && (v2==v4)) || ((v1==v4) && (v2==v3)) ) stop=1;
          else k++;
        }
	//printf("FIN while ((k<=ur[0]->NbElem()) && (!stop)) j=%d\n",j);
	
	// First check if the objects in question has already been 
	// deleted or not.
	if ((DB->GetObj(v1) != NULL) && (DB->GetObj(v2) != NULL)) {
	  if ((!stop) &&(((OPPORTUNISTIC_READ == 0) 
			  || ((buffmgr->ObjInMem(v1) == true)
			      && (buffmgr->ObjInMem(v2) == true)))
			 && ((OPPORTUNISTIC_WRITE == 0) ||
			     ((buffmgr->ObjInUpdatedMem(v1) == true)
			      && (buffmgr->ObjInUpdatedMem(v2) == true))))) {
	    client->o1->Insert(place,v1);
	    client->o2->Insert(place,v2);
	    client->fc->Insert(place,DB->mc[v1][1]->Scan(j));
	    place++;
	  }
	}
      }
    }
    //printf("FIN for (j=1; j<=DB->mc[v1][0]->NbElem(); j++) i=%d\n",i);
  }  
  //printf("FIN for (i=1; i<=n; i++)\n");
  /*
  printf("o1 ");client->o1->Display();
  printf("o2 ");client->o2->Display();
  printf("fc ");client->fc->Display();  */
  
  // Constitution des UR
  
  
  while (client->o1->NbElem()>0) {
    // Creation de l'UR courante (liste d'OIDs)
    ur[0]=new ChList;
    ur[1]=new ChList;
    ur[2]=new ChList;
    // Premier couple (o1,o2) "depile" dans l'UR courante
    ur[0]->Insert(1,client->o1->Scan(1));
    client->o1->Delete(1);
    ur[1]->Insert(1,client->o2->Scan(1));
    client->o2->Delete(1);
    ur[2]->Insert(1,client->fc->Scan(1));
    client->fc->Delete(1);
    // Enrichissement de l'UR courante a partir de (o1,o2)
    stop=0;
    while ((!stop) && (client->o1->NbElem()>0)) {
      // Parcours de (o1,o2)
      stop=1;
      place=1;
      while (place<=client->o1->NbElem()) {
        v1=client->o1->Scan(place);
        v2=client->o2->Scan(place);
        // printf("(o1,o2)=(%d,%d)\n",v1,v2);
        // Parcours de l'UR courante pour trouver les composantes connexes
        stop2=0;
        place2=1;
        while ((place2<=ur[0]->NbElem()) && (!stop2)) {
          v3=ur[0]->Scan(place2);
          v4=ur[1]->Scan(place2);
          // printf("(ur1,ur2)=(%d,%d)\n",v3,v4);
          if ((v1==v3) || (v2==v3) || (v1==v4) || (v2==v4)) {
            // Composante connexe detectee
            stop2=1;
            stop=0;
            // Insertion dans l'UR courante
            ur[0]->Insert(1,v1);
            ur[1]->Insert(1,v2);
            ur[2]->Insert(1,client->fc->Scan(place));
            // Couple efface de (o1,o2)
            client->o1->Delete(place);
            client->o2->Delete(place);
            client->fc->Delete(place);
            place--;
          } else place2++;
        }
        place++;
      }
    } // Fin while ((!stop) && (client->o1->NbElem()>0)) {

    /*
    printf("Current UR: \n");
    ur[0]->Display();
    ur[1]->Display();
    ur[2]->Display();   
    */
    // hack attempt at bug fix
  
    int ur_v0;
    int ur_v1;
  
    bool found_delete = true;
    while (found_delete == true) {
      found_delete = false;
      // duplicate elminiation
      for (int zhen=1; (zhen <= (ur[0]->NbElem())); zhen++) {
	ur_v0 = ur[0]->Scan(zhen);
	ur_v1 = ur[1]->Scan(zhen);
	for (int he=1; he <= ur[0]->NbElem(); he++) {
	  if ((zhen != he) && ((ur_v0 == ur[0]->Scan(he)) &&
			       (ur_v1 == ur[1]->Scan(he)))) {
	    ur[0]->Delete(he);
	    ur[1]->Delete(he);
	    ur[2]->Delete(he);
	    found_delete = true;
	    break;
	  }
	}
	if (found_delete == true) {
	  break;
	}
      }
    }
    
    /*
    printf("After dup elem UR: \n");
    ur[0]->Display();
    ur[1]->Display();
    ur[2]->Display();   
    */

    // Ordonnancement de l'UR courante => cluster en sortie

    cluster=Schedule(ur);


//    printf("Current cluster ");cluster->Display();
    NCLUSTERS++;
    TCLUSTERS+=cluster->NbElem();


    // Code used to set clustered together flag

    //  cout << "cluster size = " << cluster->NbElem() << endl;
    for (i=1; i<=cluster->NbElem(); i++) { 
      set<int> objects_flag_set;
      obj=cluster->Scan(i);      
      for (int inner_i = 1; inner_i <= cluster->NbElem(); inner_i++) { 
	if (i != inner_i) {
	  int place_flag=1;
	  int stop_flag=0;
	  int inner_obj = cluster->Scan(inner_i);
	  while ((place_flag<=DB->mc[obj][0]->NbElem()) && (!stop_flag))
	    if (DB->mc[obj][0]->Scan(place_flag)<inner_obj) place_flag++;
	    else stop_flag=1;
	  if (stop_flag)
	    if (DB->mc[obj][0]->Scan(place_flag)==inner_obj) {
	      DB->mc[obj][3]->Update(place_flag,1);
	      objects_flag_set.insert(inner_obj);
	    }
	}
      }
      for (int inner_i = 1; inner_i <= DB->mc[obj][0]->NbElem(); inner_i++) {
	int inner_obj = DB->mc[obj][0]->Scan(inner_i);
	if ((inner_obj != obj) &&
	    (objects_flag_set.find(inner_obj) == objects_flag_set.end())) {
	  DB->mc[obj][3]->Update(inner_i, 0); 
	}
      }
    }

    // Placement du cluster sur disque

    // Creation nouvelle page
    newpage=DB->NewPage();
    DB->AddPage(newpage);
    // Placement de tous les objets du cluster
    for (i=1; i<=cluster->NbElem(); i++) {
      obj=cluster->Scan(i);
      // Chargement de l'objet courant
      client->rootoid=obj;
      // cout << "obj = " << obj << endl;
     
      time+=ObjRequest(client);
      // Objet courant efface de la (ou des) page(s) ou il est stocke
//      printf("Placement of object #%d\n",obj);
      while (DB->GetObj(obj)->PageDisk()->NbElem()>0) {
        // On depile
        v1=DB->GetObj(obj)->PageDisk()->Scan(1);
        cpage=DB->GetPage(v1);
        DB->GetObj(obj)->PageDisk()->Delete(1);
        // Suppression de l'objet dans la page cpage/v1
        if (cpage!=NULL) {
          rsize=cpage->Free();
          j=1;
          stop=0;
          while ((j<=cpage->Objects()->NbElem()) && (!stop))
            if (cpage->Objects()->Scan(j)==obj) stop=1;
            else j++;
          if (stop) {
            // Suppression
            cpage->Objects()->Delete(j);
            // Evolution taille restante
	    rsize+=DB->GetObj(obj)->ClassPtr()->GetSize();
          }
//          printf(" Object #%d removed from page #%d, free space=%d\n",obj,v1,rsize);
	  // If all the objects from the source has been extracted then
	  // delete the source page.
	  if (rsize<PGSIZE) // Maj place libre
            cpage->ResetFree(rsize);
	  else // Page vide, donc effacee
	    DB->DelPage(v1);
        } // Fin if (cpage!=NULL)
        // Flag de mise a jour de la page v1 mis a 1 dans le buffer
        buffmgr->SetUpdated(v1);
      } // Fin while (DB->GetObj(obj)->PageDisk()->NbElem()>0)
      // Placement de l'objet courant
      // put in place the actual object
      if (DB->GetObj(obj)->ClassPtr()->GetSize()<=PGSIZE) {
        // Objet normal de taille inferieure a celle d'une page
        if (DB->GetObj(obj)->ClassPtr()->GetSize()>newpage->Free()) { // Taille > place libre
          // Sauvegarde page precedente
	  // save previous page
          time+=AccDisk(newpage->PageNum());
	  buffmgr->SetUpdated(newpage->PageNum());
          // Nouvelle page
	  // new page
          newpage=DB->NewPage();
          DB->AddPage(newpage);
        }
        // Objet insere dans le contenu de la page
        newpage->Objects()->Insert(1,obj);
        // Page placee dans la liste des pages de l'objet
        DB->GetObj(obj)->PageDisk()->Insert(1,newpage->PageNum());
        // M.a.j. place libre
        newpage->ResetFree(newpage->Free()-DB->GetObj(obj)->ClassPtr()->GetSize());
	// printf("Object #%d moved to page #%d\n",obj,newpage->PageNum());
      } else { // Objet long
        // rsize : taille restant a placer
        rsize=DB->GetObj(obj)->ClassPtr()->GetSize()-newpage->Free();
        // Page courante saturee
        newpage->ResetFree(0);
        // Objet insere dans le contenu de la page
        newpage->Objects()->Insert(1,obj);
        // Page dans la liste des pages de l'objet
        DB->GetObj(obj)->PageDisk()->Insert(1,newpage->PageNum());
	// printf("Object #%d moved to page #%d\n",obj,newpage->PageNum());
        // Tant que l'objet n'est pas totalement stocke
        while (rsize>0) {
          // Nouvelle page
          newpage=DB->NewPage();
          DB->AddPage(newpage);
          // Objet insere dans le contenu de la page
          newpage->Objects()->Insert(1,obj);
          // Page dans la liste des pages de l'objet
          DB->GetObj(obj)->PageDisk()->Insert(1,newpage->PageNum());
          if (newpage->Free()<rsize) {   // L'objet ne tient pas
            rsize=rsize-newpage->Free(); // Taille restante --
            newpage->ResetFree(0);       // Page saturee
          } else {
            newpage->ResetFree(newpage->Free()-rsize); // Espace libre --
            rsize=0;
          }
	  // printf("Object #%d moved to page #%d\n",obj,newpage->PageNum());
        }
      }
    } // Fin for (i=1; i<=cluster->NbElem();
    // Sauvegarde derniere page
    time+=AccDisk(newpage->PageNum());
    buffmgr->SetUpdated(newpage->PageNum());

    // Destruction UR courante et cluster courant

    delete ur[0];
    delete ur[1];
    delete ur[2];
    delete cluster;

  } // Fin while (client->o1->NbElem()>0)
  //exit(1);

  // Destruction des listes de regroupement

  delete client->o1;
  delete client->o2;
  delete client->fc;

  return time;

} // Fin de procedure

float ClustMgr::DROAnalysis(Client * client) {
  float time;
  vector<DROPageStats *> clustPages;
  DROPageStats * pageStats; 
  float pageUR;
  int pageLNum;
  int numPagesUsed = 0;
  float clustRatio;

  time = 0.0;  
  
  hash_map<int, DROPageStats *, hash<int>, eqoid>::const_iterator it;
  for (it = DROPageStatsMap.begin(); it != DROPageStatsMap.end(); it++) {
    
    pageStats = it->second;
    pageUR = pageStats->computeUsageRate();
    pageLNum = pageStats->getNbLoad();

    if ( (pageUR > DROMINUR) && (pageLNum > DROMINLT) ) {
      clustPages.push_back(pageStats);
    }

    // number of pages actually used
    if (pageUR > 0.0) {
      numPagesUsed++;
    }

  }

  if (numPagesUsed > 0 ) {
    clustRatio = (float) clustPages.size() / (float) numPagesUsed;
    
    if ((clustPages.size() > 1) && ( clustRatio > DROPCRate)) {
      
      time += DROReorg(clustPages, client);

    }
  }

  return time;
}

float ClustMgr::DROReorg(vector<DROPageStats *> & clustPages, Client * client) {

  float time;
  PQueue clustObjPQ;
  hash_map<int, int, hash<int>, eqoid> clustFreqMap;
  DROPageStats * pageStats;
  vector<int> tempVect;
  vector<int> clustList;
  int currOid;
  int currFreq;
  float resembRate;

  time  = 0.0;
  
  for( unsigned int i = 0; i < clustPages.size(); i++ ) {

    pageStats = clustPages[i];
    tempVect.clear();
    tempVect = pageStats->getUsedObjects();

    for (unsigned int j = 0; j < tempVect.size(); j++) {
      
      currOid = tempVect[j];
      currFreq = pageStats->getObjAccessFreq(currOid);
      clustObjPQ.InsertElem((float) currFreq, currOid);
      clustFreqMap[currOid] = currFreq;
      //      cout << "obj = " << currOid << ", freq = " << currFreq << endl;
      // cout << "inserting object = " << currOid << endl;

      //  cout << "obj size = " << DB->GetObj(currOid)->ClassPtr()->GetSize();
      //      cout << endl;

    }
    
  } 
  
  time += DROGenClustList(clustObjPQ, clustFreqMap, clustList);
 
  resembRate = DROCompResembRate(clustList, clustPages);
  if (resembRate < DROMaxRR) {
    
    vector<int> movedPages;    
    time += DROPhysicalReorg(clustList, clustPages, movedPages, client);
    time += DROStatsUpdate(movedPages);
    
  }

  return time;
}

float ClustMgr::DROStatsUpdate(vector<int> & movedPages) {

  float time;
  int currPid, currOid;
  vector<int> movedObjects;
  vector<int> tempVect;
  DROPageStats * pageStats;
  hash_map<int, DROObjStats *, hash<int>, eqoid> objStatsMap;
  vector<DROObjStats *> tempStatsVect;
  DROObjStats * tempStats;
  
  time = 0.0;
  
  for ( unsigned int i = 0; i < movedPages.size(); i++ ) {

    currPid = movedPages[i];
    pageStats = DROPageStatsMap[currPid];
    tempVect = pageStats->getAllObjects();

    for (unsigned int j = 0; j < tempVect.size(); j++) {

      movedObjects.push_back(tempVect[j]);      

    }

    pageStats->resetNbLoad();   
    tempStatsVect =  pageStats->removeAllObjStats();

    for (unsigned int j = 0; j < tempStatsVect.size(); j++) {
      tempStats = tempStatsVect[j];
      tempStats->resetUsageIndicator();
      objStatsMap[tempStats->getOid()] = tempStats;

    }
      
  }  

  
  for (unsigned int i = 0; i < movedObjects.size(); i++) {

    currOid = movedObjects[i];
    currPid = DB->GetObj(currOid)->PageDisk()->Scan(1);
    if (DROPageStatsMap.find(currPid) == DROPageStatsMap.end()) {
      
      pageStats = new DROPageStats(currPid);      
      DROPageStatsMap[currPid] = pageStats;


    }
    else {

      pageStats = DROPageStatsMap[currPid];

    }

    tempStats = objStatsMap[currOid];
    pageStats->insertObjStats(tempStats);
  }

  return time;

}


struct DROSearchStruct {

  int objId;
  int depth;

};

DROClustObj * 
ClustMgr::DROResearchProcedureResult(DROClustObj * startingObject, int d,
				     hash_map<int, int, hash<int>, eqoid>
				     & clustFreqMap,
				     hash_map<int, DROClustObj *, hash<int>,
				     eqoid> & oidObjClustMap, 
				     DROClustObj * chainStartObject ) {


  int startObjFreq, currFreq, maxFreq;
  DROSearchStruct ss, ssInner;
  int startingOid;
  stack<DROSearchStruct> searchStack;
  Object * tempObj;
  DROClustObj * currClustObj;
  float dissimRate;
  int absDiff;
  int chainStartOid;

  chainStartOid = chainStartObject->getOid();
  //  cout << "DRO research procedure called" << endl;
  startingOid = startingObject->getOid();
  startObjFreq = clustFreqMap[startingOid];
  ss.objId = startingOid;
  ss.depth = 0;

  searchStack.push(ss);

  while (searchStack.empty() == false) {

    ss = searchStack.top();
    searchStack.pop();

    if ((ss.objId != startingOid) && 
	(oidObjClustMap.find(ss.objId) != oidObjClustMap.end()) &&
	(ss.objId != chainStartOid)) {      
      
      currClustObj = oidObjClustMap[ss.objId];
      if (currClustObj->getPlacementPrev() == NULL) {

	currFreq = clustFreqMap[ss.objId];
	//	cout << "startobjfreq = " << startObjFreq << endl;
	//cout << "currfreq = " << currFreq << endl;
	if (currFreq > startObjFreq) {
	  maxFreq = currFreq;
	  absDiff = currFreq - startObjFreq;
	}
	else {
	  maxFreq = startObjFreq;
	  absDiff = startObjFreq - currFreq;
	}

	dissimRate = (float) absDiff / (float) maxFreq;

	if (dissimRate < DROMaxDR) {
	  // cout << "found similar object to cluster = " << ss.objId << endl;
	  // cout << "starting obj = " << startingOid << endl;
	  return currClustObj;

	}

      }

    }

    if (ss.depth < d) {
      
      for (int i  = 0; i < DB->GetObj(ss.objId)->ClassPtr()->MaxNRef(); i++) {
	tempObj = DB->GetObj(ss.objId)->GetORef(i);
	if (tempObj != NULL) {
	  //  cout << "child oid = " << tempObj->OID() << endl;
	  ssInner.objId = tempObj->OID();
	  ssInner.depth = ss.depth + 1;
	  searchStack.push(ssInner);
	}
      }
    }
  }

  return NULL;
}


// uses algorithm specified in DRO paper page 78 of International symposium
// on objects and databases in Sophia Antipolis France June 2000

float
ClustMgr::DROGenClustList(PQueue & clustObjPQ, 
			  hash_map<int, int, hash<int>, eqoid> & clustFreqMap,
			  vector<int> & clustList) {
  
  int d = 0;
  //  bool end = false;
  DROClustObj * startingObject, * objectToLink, * foundObject, * clustFirstObj;
  int currOid;
  DROClustObj * prevClustObj, * currClustObj;
  bool found;
  hash_map<int, DROClustObj *, hash<int>, eqoid> oidObjClustMap;
  float time;
  DROClustObj * currStartObject;
  set<int> tempCheckSet;
  time = 0.0;

  //  cout << "clustObjPQ size = " << clustObjPQ.Size() << endl;

  currOid = clustObjPQ.StartIter();
  clustFirstObj = new DROClustObj(currOid);
  prevClustObj = clustFirstObj;
  oidObjClustMap[currOid] = clustFirstObj;

  //  cout << "PQ freq = " << clustFreqMap[currOid] << endl;
  // cout << "oid = " << currOid  << endl;

  currOid = clustObjPQ.Next();
  while (currOid != clustObjPQ.End()) {
    //  cout << "PQ freq = " << clustFreqMap[currOid] << endl;
    // cout << "oid = " << currOid  << endl;
    currClustObj = new DROClustObj(currOid);
    oidObjClustMap[currOid] = currClustObj;
        
    prevClustObj->setSortNext(currClustObj);
    prevClustObj = currClustObj;
    
    currOid = clustObjPQ.Next();
  }

  // this was used in the paper but I can not see the need for the variable end
  // also end is never set to true in the paper.

  // while ( (d < MaxD) && (end == false) ) {

  while  (d < DROMaxD) {
    d++;

    // Browse objects to cluster
    startingObject = clustFirstObj;
    
    while ((startingObject != NULL) &&
	   (startingObject->getPlacementPrev() != NULL)) {

      startingObject = startingObject->getSortNext();

    }

    currStartObject = startingObject;

    while (startingObject != NULL) {
      //      cout << "in this loop" << endl;

      objectToLink = currStartObject;

      while ((objectToLink != NULL) &&
	     (objectToLink->getPlacementNext() != NULL))  {

	objectToLink = objectToLink->getPlacementNext();

      }

      found = true;

      while (found == true) {

	// find an object to cluster different from startingObject, referenced
	// on a distance lower than d, with a dissimilarity rate lower than
	// MaxDR, and attribute placementPrev set to NULL

	foundObject = DROResearchProcedureResult(currStartObject, d,
						 clustFreqMap, oidObjClustMap,
						 startingObject);
	if (foundObject != NULL) {

	  objectToLink->setPlacementNext(foundObject);
	  foundObject->setPlacementPrev(objectToLink);
	  objectToLink = foundObject;
	  while ((objectToLink != NULL) &&
		 (objectToLink->getPlacementNext() != NULL))  {
	    
	    objectToLink = objectToLink->getPlacementNext();
	    
	  }
	  
	}
	else {

	  found = false;

	}
      }
      
      // startingObject = startingObject->getPlacementNext();

      //      while ((startingObject != NULL) && 
      //	     (startingObject->getSortNext() != NULL)) {

      //  if (startingObject->getPlacementNext() != NULL) {
      //	startingObject = startingObject->getPlacementNext();
      // }
      // else {
      //      cout << "currStartObject = " << currStartObject << endl;
      currStartObject = currStartObject->getPlacementNext();
      if (currStartObject == NULL) {
	startingObject = startingObject->getSortNext();
	//	cout << "startingObject = " << startingObject << endl;
	while ((startingObject != NULL) && 
	       (startingObject->getPlacementPrev() != NULL)) { 
	  startingObject = startingObject->getSortNext();
	  
	}
	currStartObject = startingObject;
      }
	// }
	//  }

    }

  }
  /*
  hash_set<int, hash<int>, eqoid> processedObjs;

  currOid = clustObjPQ.StartIter();


  while (currOid != clustObjPQ.End()) {
    
    currClustObj = oidObjClustMap[currOid];      

    if ((processedObjs.find(currOid) == processedObjs.end()) &&
	(currClustObj->getPlacementPrev() == NULL)) {

      while (currClustObj != NULL) {
	currOid = currClustObj->getOid();

	if (processedObjs.find(currOid) != processedObjs.end()) {
	  // cout << "Same object linked into more than one list, eventm.h";
	  //cout << endl;
	  exit(1);
	}

	processedObjs.insert(currOid);
	clustList.push_back(currOid);
	//	cout << "clust list freq = " << clustFreqMap[currOid] << endl;
	//	cout << " oid = " << currOid << endl;
	currClustObj = currClustObj->getPlacementNext();	
      }
    }    
    
    currOid = clustObjPQ.Next();
  }
  */

  vector<int> linkStartObjs;

  currOid = clustObjPQ.StartIter();

  while (currOid != clustObjPQ.End()) {
    currClustObj = oidObjClustMap[currOid];      

    if (currClustObj->getPlacementPrev() == NULL) {
      linkStartObjs.push_back(currOid);
    }   
    
    currOid = clustObjPQ.Next();
  }

  for (unsigned int i = 0; i < linkStartObjs.size(); i++) {

    currOid = linkStartObjs[i];
    currClustObj = oidObjClustMap[currOid];      

    while (currClustObj != NULL) {
      	currOid = currClustObj->getOid();
	clustList.push_back(currOid);
	currClustObj = currClustObj->getPlacementNext();	
    }

  }
  
  // delete all DROClustObj created in this procedure
  hash_map<int, DROClustObj *, hash<int>, eqoid>::iterator it;
  for (it = oidObjClustMap.begin(); it != oidObjClustMap.end(); it++) {
    
    delete it->second;

  }

  //  cout << "clust list size = " << clustList.size() << endl;

  return time;
}

float ClustMgr::DROCompResembRate( vector<int> & clustList,
				   vector<DROPageStats *> & clustPages ) {

  float resembRate;
  int currOid;

  int currPageSize;
  unsigned int count = 0;
  hash_set<int, hash<int>, eqoid> currObjSet;
  DROPageStats * pageStats;
  vector<int> usedObjs;
  int numResemb = 0;

  while (count < clustList.size()) {
    
    currPageSize = 0;
    currObjSet.clear();
    while (count < clustList.size()) {
      currOid = clustList[count];
      currPageSize += DB->GetObj(currOid)->ClassPtr()->GetSize();
      if (currPageSize < PGSIZE) {
	currObjSet.insert(currOid);
	count++;
      }
    }
    
    int maxMatches = 0;
    int numMatches;
    for (unsigned int i = 0; i < clustPages.size(); i++) {
    
      pageStats = clustPages[i];

      usedObjs = pageStats->getUsedObjects();
  
      numMatches = 0;
      for (unsigned int j = 0; j < usedObjs.size(); j++) {

	if (currObjSet.find(usedObjs[j]) != currObjSet.end()) {
	  numMatches++;
	}

      }

      if (numMatches > maxMatches) {

	maxMatches = numMatches;

      }

    }

    numResemb += maxMatches;
	   
  }

  resembRate = (float) numResemb / (float) clustList.size();

  return resembRate;
}

float ClustMgr::DROPhysicalReorg(vector<int> & clustList,
				 vector<DROPageStats *> & clustPages, 
				 vector<int> & movedPages, Client * client) {
  int currPid, currOid;
  set<int> pageSet;
  multimap<const int, part_info_t, int_cmp> obj_seq;
  vector<int> unUsedObjects;
  part_info_t pi;
  float time;
  vector<int> tempObjs;
  unsigned int totalBefore = 0;
  unsigned int totalAfter = 0;

  time = 0.0;

  for (unsigned int i = 0; i < clustPages.size(); i++) {
    
    currPid = clustPages[i]->getPid();
    movedPages.push_back(currPid);
    pageSet.insert(currPid);
    tempObjs = clustPages[i]->getAllObjects();
    totalBefore += tempObjs.size();

  }

  
  int prioNum = 0;
  set<int> * tempSet;

  for (unsigned int i = 0; i < clustPages.size(); i++) {

    unUsedObjects = clustPages[i]->getUnusedObjects();
    totalAfter += unUsedObjects.size();

    for (unsigned int j = 0; j < unUsedObjects.size(); j++) {     
      currOid = unUsedObjects[j];

      client->rootoid = currOid;
      time += ObjRequest(client);
      tempSet = new set<int>(); 
      pi.obj_set = tempSet;      
      pi.obj_set->insert(currOid);
      pi.size = DB->GetObj(currOid)->ClassPtr()->GetSize();
      obj_seq.insert(pair<int, part_info_t>(prioNum, pi));
      prioNum++;

    }

  }
  
  for (int i = ((int)clustList.size() - 1); i >= 0; i--) {
    
    currOid = clustList[i];

    totalAfter += 1;
    client->rootoid = currOid;
    time += ObjRequest(client);

    //    cout << "currOid = " << currOid << endl;
    tempSet = new set<int>(); 
    pi.obj_set = tempSet;      
    pi.obj_set->insert(currOid);
    pi.size = DB->GetObj(currOid)->ClassPtr()->GetSize();
    obj_seq.insert(pair<int, part_info_t>(prioNum, pi));
    prioNum++;

  }

  //  cout << "total objects before clustering = " << totalBefore << endl;
  //  cout << "total objects after clustering = " << totalAfter << endl;

  //  cout << "done assign" << endl;
  time += PlaceObj(&pageSet, &obj_seq, client);
  cout << "done placement" << endl;
  /*
  multimap<const int, part_info_t, int_cmp>::const_iterator obj_it;

  
  for (obj_it = obj_seq.begin(); obj_it != obj_seq.end(); obj_it++) {
    tempSet = (obj_it->second).obj_set;
        delete tempSet;
  }
  */
  return time;

}

float ClustMgr::GPCluster(Client * client) {
  float time;

  time = 0.0;
  //STATS->Display();
  //buffmgr->DisplayExtTen();
  //buffmgr->DisplayHeat();
  time += GPAnalysis(client);
  
  return time;
}


float ClustMgr::PRPCluster(Client * client) {
  float time;

  time = 0.0;
  //STATS->Display();
  //buffmgr->DisplayExtTen();
  //buffmgr->DisplayHeat();
  if (PRPCALG == 0) {
    time += PRPPRIOAnalysis(client);
  }
  else if (PRPCALG == 1) {
    time += PRPRANDAnalysis(client);
  }
  return time;
}


float ClustMgr::DROCluster(Client * client) {
  float time;

  time = 0.0;

  time += DROAnalysis(client);

  return time;

}

// CLASSE ClustMgr : Ordonnancement des UR en clusters (DSTC)

ChList *ClustMgr::Schedule(ChList *ur[3]) {

  Node *hnode;                  // Tete de la liste des noeuds
  int nodes;                    // Nombre courant de noeuds
  ChList *links[3];             // Liste des liens
                                // 0 - Noeud no 1
                                // 1 - Noeud no 2
                                // 2 - Poids
  ChList *cluster;              // Cluster en sortie
  int i, j, n, n1, n2, n3, n4;//, m;
  Node *cnode, *newnode, *prevnode;
  short found1, found2;
  int w, maxw, place;
  Node *pn1, *prevn1, *pn2, *prevn2;
  int t1, t2, q1, q2, wt1t2, wt1q2, wq1t2, wq1q2, w1, w2;
  int ndel, nkeep;

  // Creation de la liste des noeuds et de la liste des liens

  hnode=NULL;
  nodes=0;
  links[0]=new ChList;
  links[1]=new ChList;
  links[2]=new ChList;
  for (i=1; i<=ur[0]->NbElem(); i++) {
    // Parcours de la liste des noeuds pour ne pas ajouter deux fois
    // le meme objet
    cnode=hnode;
    found1=0;
    found2=0;
    //  cout << "11111" << endl;
    while (cnode!=NULL) {
      if (cnode->list->Scan(1)==ur[0]->Scan(i)) {
        found1=1;
        n1=cnode->num;
      }
      if (cnode->list->Scan(1)==ur[1]->Scan(i)) {
        found2=1;
        n2=cnode->num;
      }
      cnode=cnode->next;
    } // Fin while (cnode!=NULL)
    //  cout << "2222" << endl;
    if (!found1) {                      // Nouveau noeud
      nodes++;
      newnode=new Node;
      newnode->num=nodes;
      /* cout << "333333333" << endl;
      cout << "newnode = " << newnode << endl;
      cout << "newnode->list = " <<newnode->list << endl;*/
      newnode->list=new ChList;
      // cout << "this = " << this << endl;
      // cout << "newnode = " << newnode << endl;
      newnode->list->Insert(1,ur[0]->Scan(i));
      // cout << "got here" << endl;
      newnode->next=hnode;
      hnode=newnode;
      n1=newnode->num;
    }
    //  cout << "44444444444444" <<endl;
    if (!found2) {                      // Nouveau noeud
      nodes++;
      // cout << "1" <<endl;
      newnode=new Node;
      newnode->num=nodes;
      newnode->list=new ChList;
      newnode->list->Insert(1,ur[1]->Scan(i));
      newnode->next=hnode;
      hnode=newnode;
      n2=newnode->num;
    }
    // cout << "2" <<endl;

    // Insertion lien courant dans la liste des liens
    links[0]->Insert(1,n1);
    links[1]->Insert(1,n2);
    links[2]->Insert(1,ur[2]->Scan(i));
    // cout << "2.5" << endl;
  } // Fin for (i=1; i<=ur[0]->NbElem();
  // cout << "3" <<endl;


/*  printf("\nNodes list:\n");
  cnode=hnode;
  while (cnode!=NULL) {
    printf("%d)   ",cnode->num);
    cnode->list->Display();
    cnode=cnode->next;
    }
  printf("\nLinks list:\n");
  links[0]->Display();
  links[1]->Display();
  links[2]->Display();  
*/
  // Ordonnancement, reduction a un seul supernoeud

  while ((nodes>1)&&(links[0]->NbElem())) {
    // Recherche du poids le plus eleve dans la liste des liens
    maxw=-1;
    for (i=1; i<=links[0]->NbElem(); i++) {
      w=links[2]->Scan(i);
      if (w>maxw) maxw=w;
    }
    /*cout << "*********links[0]******" << endl;
    links[0]->Display();
    cout << "*********links[1]******" << endl;
    links[1]->Display();
    cout << "*********links[2]******" << endl;
    links[2]->Display();
    */
    // hack attempt at bug fix

    for (int zhen=1; (zhen <= (links[0]->NbElem())); zhen++) {
      if (links[0]->Scan(zhen) == links[1]->Scan(zhen)) {
	links[0]->Delete(zhen);
	links[1]->Delete(zhen);
	links[2]->Delete(zhen);
	zhen =0;
      }
    }
    /*    
    if (links[0]->Scan(1) == links[1]->Scan(1)) {
	  links[0]->Delete(1);
	  links[1]->Delete(1);
	  links[2]->Delete(1);
    }
    */
    /*cout << "after elem" << endl;
    cout << "*********links[0]******" << endl;
    links[0]->Display();
    cout << "*********links[1]******" << endl;
    links[1]->Display();
    cout << "*********links[2]******" << endl;
    links[2]->Display();

     cout <<"Display hnode ********" << endl;
     if ((hnode != NULL) &&(hnode->list != NULL))
       hnode->list->Display();
    */
    // Recherche des liens de poids max
    i=1;
    n=links[0]->NbElem();
    
    // exit(1);
    while (i<=n) {
      
      if (links[2]->Scan(i)>=maxw) { // Fusion des deux noeuds
        // Recherche des deux noeuds et de leurs objets "tete" et "queue"
        // tete -> ti, queue -> qi
        cnode=hnode;
        prevnode=NULL;
        while (cnode!=NULL) {
	  if (cnode->num==links[0]->Scan(i)) {
	    pn1=cnode;
	    prevn1=prevnode;
	    t1=cnode->list->Scan(1);
	    q1=cnode->list->Scan(cnode->list->NbElem());
	  }
	  if (cnode->num==links[1]->Scan(i)) {
	    pn2=cnode;
	    prevn2=prevnode;
	    t2=cnode->list->Scan(1);
	    q2=cnode->list->Scan(cnode->list->NbElem());
	  }
	  /*
	  if (links[0]->Scan(i) == links[1]->Scan(i)) {
	    // cout <<"i = " << i << endl;
	    // cout <<"links[0]->Scan(i) = " << links[0]->Scan(i) << endl;
	    // cout <<"Display hnode ********" << endl;
	    // cout << "cnode = " << cnode->num << endl;
	    //  cout << "cnode->list" << endl;
	    // cnode->list->Display();
	    //  hnode->list->Display();	 
	    //exit(1);
	    } */
	  prevnode=cnode;
	  cnode=cnode->next;
	  
        } // Fin while (cnode!=NULL)
        // printf("t1=%d q1=%d t2=%d q2=%d\n",t1,q1,t2,q2);
        // Recherche dans UR des poids des liens entre tous
        // les couples possibles tt, tq, qq
        wt1t2=0;
        wt1q2=0;
        wq1t2=0;
        wq1q2=0;
        for (j=1; j<=ur[0]->NbElem(); j++) {
          w1=ur[0]->Scan(j);
          w2=ur[1]->Scan(j);
          //printf("w1=%d w2=%d Bond strength=%d\n",w1,w2,ur[2]->Scan(j));
          if ( ((w1==t1) && (w2==t2)) || ((w2==t1) && (w1==t2)) )
            wt1t2=ur[2]->Scan(j);
          if ( ((w1==t1) && (w2==q2)) || ((w2==t1) && (w1==q2)) )
            wt1q2=ur[2]->Scan(j);
          if ( ((w1==q1) && (w2==t2)) || ((w2==q1) && (w1==t2)) )
            wq1t2=ur[2]->Scan(j);
          if ( ((w1==q1) && (w2==q2)) || ((w2==q1) && (w1==q2)) )
            wq1q2=ur[2]->Scan(j);
        } // Fin for (j=1; j<=ur[0]->NbElem();
        //printf("wt1t2=%d wt1q2=%d wq1t2=%d wq1q2=%d\n",wt1t2,wt1q2,wq1t2,wq1q2);
        // Fusion des noeuds en respectant les regles
        if ( (wt1t2>=wt1q2) && (wt1t2>=wq1t2) && (wt1t2>=wq1q2) ) { // t1 t2
          //printf("Fusion t1t2\n");
          // Sauvegarde no de noeud a effacer et no noeud a conserver
          ndel=links[0]->Scan(i);
          nkeep=links[1]->Scan(i);
          // Fusion
          pn2->list->Reverse();
          place=pn2->list->NbElem()+1;
	  // ******* I think bug here, infinite loop ****
          for (j=1; j<=pn1->list->NbElem(); j++) {	   
            pn2->list->Insert(place, pn1->list->Scan(j));
            place++;
	  }
          // Destruction noeud pn1
          if (prevn1!=NULL) prevn1->next=pn1->next;
          else hnode=pn1->next;
          delete pn1->list;
          delete pn1;
          nodes--;
        } else if ( (wt1q2>=wt1t2) && (wt1q2>=wq1t2) && (wt1q2>=wq1q2) ) { // t1 q2
          //printf("Fusion t1q2\n");
          // Sauvegarde no de noeud a effacer et no noeud a conserver
          ndel=links[0]->Scan(i);
          nkeep=links[1]->Scan(i);
          // Fusion
          place=pn2->list->NbElem()+1;
          for (j=1; j<=pn1->list->NbElem(); j++) {
	    /*cout << "trapped" << endl;
	    if (pn1 == pn2) {
	      cout << "not good" << endl;
	      exit(1);
	      }*/
            pn2->list->Insert(place,pn1->list->Scan(j));
            place++;
          }
          // Destruction noeud pn1
          if (prevn1!=NULL) prevn1->next=pn1->next;
          else hnode=pn1->next;
          delete pn1->list;
          delete pn1;
          nodes--;
        } else if ( (wq1t2>=wt1t2) && (wq1t2>=wt1q2) && (wq1t2>=wq1q2) ) { // q1 t2
          //printf("Fusion q1t2\n");
          // Sauvegarde no de noeud a effacer et no noeud a conserver
          ndel=links[1]->Scan(i);
          nkeep=links[0]->Scan(i);
          // Fusion
          place=pn1->list->NbElem()+1;
          for (j=1; j<=pn2->list->NbElem(); j++) {
	    
            pn1->list->Insert(place,pn2->list->Scan(j));
            place++;
          }
          // Destruction noeud pn2
          if (prevn2!=NULL) prevn2->next=pn2->next;
          else hnode=pn2->next;
          delete pn2->list;
          delete pn2;
          nodes--;
        } else { // q1 q2
          //printf("Fusion q1q2\n");
          // Sauvegarde no de noeud a effacer et no noeud a conserver
          ndel=links[1]->Scan(i);
          nkeep=links[0]->Scan(i);
          // Fusion
          pn2->list->Reverse();
          place=pn1->list->NbElem()+1;
          for (j=1; j<=pn2->list->NbElem(); j++) {
	    
            pn1->list->Insert(place,pn2->list->Scan(j));
            place++;
          }
          // Destruction noeud pn2
          if (prevn2!=NULL) prevn2->next=pn2->next;
          else hnode=pn2->next;
          delete pn2->list;
          delete pn2;
          nodes--;
        }
        // Destruction lien courant
        links[0]->Delete(i);
        links[1]->Delete(i);
        links[2]->Delete(i);
        i--;

        // M.a.j. des liens contenant le noeud efface

        for (j=1; j<=links[0]->NbElem(); j++) {
	 
          if (links[0]->Scan(j)==ndel) links[0]->Update(j,nkeep);
          if (links[1]->Scan(j)==ndel) links[1]->Update(j,nkeep);
        }

/*      Solution plus couteuse en temps
        int m;
        j=1;
        m=links[0]->NbElem();
        while (j<=m) {
          if (links[0]->Scan(j)==ndel) links[0]->Update(j,nkeep);
          if (links[1]->Scan(j)==ndel) links[1]->Update(j,nkeep);
          if (links[0]->Scan(j)==links[1]->Scan(j)) {
            links[0]->Delete(j);
            links[1]->Delete(j);
            links[2]->Delete(j);
            if (j<=i) i--;
            j--;
          }
          m=links[0]->NbElem();
          j++;
        }     */


/*      printf("\nNodes list after fusion:\n");
      cnode=hnode;
      while (cnode!=NULL) {
        printf("%d)   ",cnode->num);
        cnode->list->Display();
        cnode=cnode->next;
      }
      printf("\nLinks list after fusion:\n");
      links[0]->Display();
      links[1]->Display();
      links[2]->Display();      */

      //////
      i=n+1; // STOP
      //////
      } // Fin if (links[3]->Scan(i)>=wmax)
      i++;
      n=links[0]->NbElem();
    } // Fin while (i<=n)
    // Elimination des liens redondants
    i=1;
    n=links[0]->NbElem();
    while (i<=n) {
      
      n1=links[0]->Scan(i);
      n2=links[1]->Scan(i);
      w1=links[2]->Scan(i);
      j=i+1;
      while (j<=n) { 

        n3=links[0]->Scan(j);
        n4=links[1]->Scan(j);
        if ( ((n1==n3) && (n2==n4)) || ((n1==n4) && (n2==n3)) ) {
          w2=links[2]->Scan(j);
          // Suppression du lien de poids le plus faible
          if (w1<w2) {
            links[0]->Delete(i);
            links[1]->Delete(i);
            links[2]->Delete(i);
            i--;
            j=n+1;
          } else {
            links[0]->Delete(j);
            links[1]->Delete(j);
            links[2]->Delete(j);
            j--;
            i++;
          }
        } // Fin gros if
        j++;
        n=links[0]->NbElem();
      } // Fin while (j<=n)
      i++;
      n=links[0]->NbElem();
    } // Fin while (i<=n)
/*    printf("\nConsistent links list:\n");
    links[0]->Display();
    links[1]->Display();
    links[2]->Display();
    printf("Nodes remaining: %d\n",nodes);      */
  } // Fin while (nodes>1)

  // Copie du supernoeud final dans le cluster

  cluster=new ChList;
  // cout <<"k1.01" << endl;
  for (i=1; i<=hnode->list->NbElem(); i++) {
    cluster->Insert(1,hnode->list->Scan(i));
  }

  //  cout <<"cluster list ***** "<< endl;
  // cluster->Display();
  // exit(1);
  // cout << "k1.1" << endl; 
  // Destruction de la liste des noeuds
  // (la tete de liste pointe sur le dernier supernoeud)

  delete hnode->list;
  delete hnode;

  // Destruction de la liste des liens

  delete links[0];
  delete links[1];
  delete links[2];

  // Retour du pointeur cluster
  //  cout <<"k2" << endl;
  return cluster;

} // Fin de procedure

// CLASSE ClustMgr : Requete d'objet, retourne le temps necessaire

float ClustMgr::ObjRequest(Client *client) {

  float time;
  int i, n;

  time = 0.0;

  // R3 : extraction des pages contenant l'objet client->rootoid
  objmgr->ExtPg(client);
  // Prechargement (enrichissement de la liste pgnums du client)
  objmgr->Prefetch(client);

  // R4 : Verification pages en memoire
  // This method deletes client->pgnums if the pages are in memory otherwise
  // it leaves it for us to fetch from disk through AccDisk
  buffmgr->AccPage(client);

  // Acces disque sur les pages a charger
  for (i=1; i<=client->pgnums->NbElem(); i++)
    time+=AccDisk(client->pgnums->Scan(i));
  // Insertion de toutes les pages accedees dans le cache
  for (i=1; i<=client->pgnums->NbElem(); i++) {
    n=buffmgr->Replace();             // Trouve une place (n)
    if (buffmgr->Updated(n)) // Page mise a jour => ecriture
      time+=AccDisk(buffmgr->Page(n));
    // La nouvelle page est inseree dans le cache en place n
    buffmgr->SetBuffer(n,client->pgnums->Scan(i),Sim()->Tnow()+time);
  }
  delete client->pgnums;        // Liberation memoire
  time+=buffmgr->Tnet();        // Transfert reseau BuffMgr->ObjMgr

  // Transfert reseau ObjMgr -> ClustMgr
  time+=objmgr->Tnet()*DB->GetObj(client->rootoid)->ClassPtr()->GetSize();

  return time;
}

// CLASSE ClustMgr : Acces disque direct, retourne le temps necessaire

float ClustMgr::AccDisk(int pagenum) {

  float time;
  

  time=DISKLAT+DISKTRA;
  if ((pagenum<(lastpage-1)) || (pagenum>(lastpage+1))) time+=DISKSEA; // Acces non-contigu
  lastpage=pagenum;
  NBIO++;
  CNBIO++;
  return time;
}

/////////////////////////////////////////////////////////////////////
// CLASSE ObjMgr (RA4)
/////////////////////////////////////////////////////////////////////

// CLASSE ObjMgr : Constructeur

ObjMgr::ObjMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t): Resource(n, cap, sim) {

  rsproc[0]=rsp;
  tnet=t;
  _simul_upper = sim;
}

void ObjMgr::SetBuffPtr(BuffMgr * bm) {
  buffmgr = bm;
}

// CLASSE ObjMgr : Retourne tnet

float ObjMgr::Tnet() {

  return tnet;
}

// CLASSE ObjMgr : Evenement 0 - Reservation RA4

void ObjMgr::Event40(Client *client) {

  this->P(41,client,1);
}

// CLASSE ObjMgr : Evenement 1 - Demande processeur

void ObjMgr::Event41(Client *client) {

  rsproc[client->cliid]->P(42,client,1);
}

// CLASSE ObjMgr : Evenement 2 - R3 + prechargement

void ObjMgr::Event42(Client *client) {
  
  // R3
  ExtPg(client);
  // Prefetching (enrich client pgnums list)
  Prefetch(client);
  // Release processor
  rsproc[client->cliid]->V();
  // Page request
  Sim()->Sched()->Schedule(50,Sim()->Tnow(),client); // Request transmitted
}

// CLASSE ObjMgr : Evenement 3 - Objet en memoire

void ObjMgr::Event43(Client *client) {

  this->V();                            // Liberation RA4
  // If the system is DB server SYSCLASS 4 then include network overhead of
  // returning object to client
  if (SYSCLASS == 4) {
    Sim()->Sched()->Schedule(26, Sim()->Tnow()+tnet*DB->GetObj(client->rootoid)->ClassPtr()->GetSize(),client);
  }
  else {
    Sim()->Sched()->Schedule(26, Sim()->Tnow(), client);
  }
  // Statistiques nombre d'objets accedes
  NBOBJ[client->type-1]++;
  NBOBJ[NTTRANS]++;
  _MoveHotRegion();
  
}

// Shift the hot region of the benchmark if it is time to.  The hot region is
// moved every HOT_REGION_MOVE_RATE object accesses.
void ObjMgr::_MoveHotRegion() {
  
  if ((HOT_REGION_MOVE_RATE != 0.0) && (HOT_REGION_MOVE_SIZE !=0.0)) {
    if ((NBOBJ[NTTRANS] % HOT_REGION_MOVE_RATE) == 0) {
      if (HOT_MOVE_FORWARD == true) {
	if ((HOT_START + HOT_REGION_MOVE_SIZE) >= (float)(1.0 - HOTRANGE)) {
	  HOT_START = HOT_START - HOT_REGION_MOVE_SIZE;
	  HOT_MOVE_FORWARD = false;
	}
	else {
	  HOT_START = HOT_START + HOT_REGION_MOVE_SIZE;
	}
      }
      else {
	if ((HOT_START - HOT_REGION_MOVE_SIZE) < 0.0) {
	  HOT_START = HOT_START + HOT_REGION_MOVE_SIZE;
	  HOT_MOVE_FORWARD = true;
	}
	else {
	  HOT_START = HOT_START - HOT_REGION_MOVE_SIZE;
	}
      }
    }
  }
}

// CLASSE ObjMgr : R3 - Extraire page

void ObjMgr::ExtPg(Client *client) {

  int i, oid;

  client->pgnums=new ChList();
  // The object could be an large object in which case it would span multiple
  // pages
  for (i=1; i<=DB->GetObj(client->rootoid)->PageDisk()->NbElem(); i++) {
    oid=DB->GetObj(client->rootoid)->PageDisk()->Scan(i);
    client->pgnums->Insert(1,oid);
  }

//  printf("Object #%d page ",client->rootoid);
//  client->pgnums->Display();
}

// CLASSE ObjMgr : Routine de prechargement
// Precondition  : The root object and its referenced objects are smaller than
//                 a page in size.
void ObjMgr::Prefetch(Client *client) {
 
  if (PREFETCH==9) {                    // User-defined strategy
    // Insert code here
    // Enrich client->pgnums list with the page to access
  }
  // Else no treatment
//  printf("[After prefetching] Object #%d page ",client->rootoid);
//  client->pgnums->Display();
}

/////////////////////////////////////////////////////////////////////
// CLASSE BuffMgr (RA5)
/////////////////////////////////////////////////////////////////////

// CLASSE BuffMgr : Constructeur

BuffMgr::BuffMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t): Resource(n, cap, sim) {

  rsproc=rsp;
  tnet=t;
  _wss_num_pages = 0;
  _num_wins_itered = 0;
  _total_distinct_pages = 0;
  _wss_distinct_pages = 0;
}


int BuffMgr::GetTotalDistinctPages() {
  return _total_distinct_pages;
}

int BuffMgr::GetNumWinsItered() {
  return _num_wins_itered;
}

int BuffMgr::GetLTEFNumDPages() {
  return _ltef_distinct_pages.size();
}

// CLASSE BuffMgr : Retourne tnet

float BuffMgr::Tnet() {

  return tnet;
}

// CLASSE BuffMgr : Retourne 1 si la page num a ete "mise a jour", 0 sinon

short BuffMgr::Updated(int num) {

  if (num<=buffpos) return buffer[num].updated;
  else return 0;
}

// CLASSE BuffMgr : Retourne le numero de la page en position num

int BuffMgr::Page(int num) {

  if (num<=buffpos) return buffer[num].pgnum;
  else return -1;
}

bool BuffMgr::PageInMem(int page_num) {
  bool found = false;
  
  for(int i=0; i < BUFFSIZE; i++) {
    if (buffer[i].pgnum == page_num) {
      found = true;      
      break;
    }
  }
  

  return found;
}


bool BuffMgr::NonEmptyPageInMem(int page_num) {
  bool found = false;
  
  for(int i=0; i < BUFFSIZE; i++) {
    if (buffer[i].pgnum == page_num) {
      found = true;      
      break;
    }
  }
  
  if (found == true) {
    found = !DB->PageEmpty(page_num);
  }

  return found;
}

bool BuffMgr::UpdatedPageInMem(int page_num) {
  bool found = false;
  
  for(int i=0; i < BUFFSIZE; i++) {
    if ((buffer[i].pgnum == page_num) && (buffer[i].updated == 1)) {
      found = true;      
      break;
    }
  }

  return found;
}

bool BuffMgr::NonEmptyUpdatedPageInMem(int page_num) {
  bool found = false;
  
  for(int i=0; i < BUFFSIZE; i++) {
    if ((buffer[i].pgnum == page_num) && (buffer[i].updated == 1)) {
      found = true;      
      break;
    }
  }

  if (found == true) {
    found = !DB->PageEmpty(page_num);
  }
  return found;
}



bool BuffMgr::ObjInMem(int obj_id) {
  bool found = false;
  int page_num;
  Object * obj_ptr;
  
  obj_ptr = DB->GetObj(obj_id);
  if (obj_ptr != NULL) {
    page_num = obj_ptr->PageDisk()->Scan(1);
    for(int i=0; i < BUFFSIZE; i++) {
      if (buffer[i].pgnum == page_num) {
	found = true;      
	break;
      }
    }
  }
  else {
    cerr << "ERROR: object checked have already been removed, eventm.h" << endl;
    return false;
  }

  return found;
}

bool BuffMgr::ObjInUpdatedMem(int obj_id) {
  bool found = false;
  int page_num;
  Object * obj_ptr;
  
  obj_ptr = DB->GetObj(obj_id);
  if (obj_ptr != NULL) {
    page_num = obj_ptr->PageDisk()->Scan(1);
    for(int i=0; i < BUFFSIZE; i++) {
      if ((buffer[i].pgnum == page_num) && (buffer[i].updated == 1)) {
	found = true;      
	break;
      }
    }
  }
  else {
    return false;
  }

  return found;
}

// CLASSE BuffMgr : Initialisation du cache a vide

void BuffMgr::InitBuffer() {

  int i, j;

  for (i=0; i<BUFFSIZE; i++) {
    buffer[i].pgnum=-1;
    buffer[i].updated=0;
    buffer[i].nbacc=0;    
    for (j=0; j<KLRU; j++) buffer[i].date[j]=0;
    buffer[i].clockflag=0;
  }

  buffpos=-1;
  clock=0;
}

// CLASSE BuffMgr : Marque la page numero num a "mise a jour"
// Takes page number of page to be updated as input parameter
void BuffMgr::SetUpdated(int num) {

  int i;

  i=0;
  while (i<=buffpos) {
    if (buffer[i].pgnum==num) {
      buffer[i].updated=1;
      i=buffpos+1;
    } else i++;
  }
}

// CLASSE BuffMgr : Insere la page pgnum en place num

void BuffMgr::SetBuffer(int num,int pgnum,float time) {

  int j;

  buffer[num].pgnum=pgnum;
  buffer[num].updated=0;
  buffer[num].nbacc=1;

  for (j=0; j<KLRU; j++) buffer[num].date[j]=time;


  if (PGREP == 7) {
    Dates * newDates = new Dates;
    
    for (j = 1; j < KLRU; j++) { 
      newDates->date[j] = 0;
    }
    
    newDates->date[0] = time;
    LRU_K_Hist[pgnum] = newDates;
    LRU_K_Last[pgnum] = time;
  }


  buffer[num].clockflag=1;
}

// CLASSE BuffMgr : Evenement 0 - Reservation RA5

void BuffMgr::Event50(Client *client) {

  this->P(51,client,1);
}

// CLASSE BuffMgr : Evenement 1 - Demande processeur (1)

void BuffMgr::Event51(Client *client) {

  rsproc->P(52,client,1);
}

// CLASSE BuffMgr : Evenement 2 - R4 et requete d'E/S si necessaire

void BuffMgr::Event52(Client *client) {
 int curr_pg;
 bool found_dup;

 // calculation of working set size
 for (int i = 1; i <= client->pgnums->NbElem(); i++) {
   if (_wss_num_pages == WSS_WIN_SIZE) {
     _wss_num_pages = 0;
     _wss_window.clear();
     _num_wins_itered++;
     _total_distinct_pages = _total_distinct_pages +  _wss_distinct_pages;
     _wss_distinct_pages = 0;
   }
   
   curr_pg = client->pgnums->Scan(i);    
   found_dup = false;
   for (unsigned int j = 0; j < _wss_window.size(); j++) {
     if (_wss_window[j] == curr_pg) {
       found_dup = true;
       break;
     }
   }
   if (found_dup == false) {
     _wss_distinct_pages++;
     _wss_window.push_back(curr_pg);
   }
   _wss_num_pages++; 
 }
 
 if (NBTRANS[NTTRANS] > LT_EF_ON_NUM_TRANS) {
   // for calcuation of long term expansion factor
   for (int i = 1; i <= client->pgnums->NbElem(); i++) {
     curr_pg = client->pgnums->Scan(i);    	
     _ltef_distinct_pages.insert(curr_pg);
   }    
  }
 
 AccPage(client);              // R4
 rsproc->V();                  // Liberation processeur
 // Les pages restantes dans pgnums sont a charger
 if (client->pgnums->NbElem()) { // if there are some pages that need
   // to be loaded.
   Sim()->Sched()->Schedule(60,Sim()->Tnow(),client);   
 }
 else  { // Rien a charger    
   Sim()->Sched()->Schedule(55, Sim()->Tnow() ,client);
 }
}

// CLASSE BuffMgr : Evenement 3 - Demande processeur (2)

void BuffMgr::Event53(Client *client) {

  rsproc->P(54,client,1);
}

// CLASSE BuffMgr : Evenement 4 - Remplacement de page

void BuffMgr::Event54(Client *client) {

  int i, j, n;
  Client *newcli;

  
  // Insertion de toutes les pages accedees dans le cache
  for (i=1; i<=client->pgnums->NbElem(); i++) {
    n=Replace();             // Trouve une place (n)
    
    // reset object usage indicator for DRO
    //DROPageStatsMap[buffer[n].pgnum]->resetAllObjUsageIndicator();

    if (buffer[n].updated) { // Page mise a jour => ecriture
      newcli=Sim()->NewClient();
      newcli->type=DDWT;
      WNBIO++;
      newcli->pgnums=new ChList;
      newcli->pgnums->Insert(1,buffer[n].pgnum);
      Sim()->Sched()->Schedule(60,Sim()->Tnow(),newcli); // Requete ecriture envoyee
    }
    // La nouvelle page est inseree dans le cache en place n
    buffer[n].pgnum=client->pgnums->Scan(i);
    if (client->update_root_obj == true) {
      buffer[n].updated=1;
    }
    else {
      buffer[n].updated=0;
    }

    buffer[n].nbacc = 1;

    for (j=0; j<KLRU; j++) buffer[n].date[j]=Sim()->Tnow();

    
    if (PGREP == 7) {
      Dates * tempDates; 
      int pgnum = buffer[n].pgnum;
      
#ifdef LATE_DEBUG
      cout << "inserted pg num = " << pgnum << endl;
#endif
      if (LRU_K_Hist.find(pgnum) == LRU_K_Hist.end()) {
	tempDates = new Dates;
	for (j = 1; j < KLRU; j++)       
	  tempDates->date[j] = 0; 
	
	LRU_K_Hist[pgnum] = tempDates;
      }
      else {
	tempDates = LRU_K_Hist[pgnum];
	for (j = 1; j < KLRU; j++) 
	  tempDates->date[j] = tempDates->date[j-1];
      }
      
      LRU_K_Hist[pgnum]->date[0] = Sim()->Tnow();
      LRU_K_Last[pgnum] = Sim()->Tnow();
    }
    buffer[n].clockflag=1;

  }
  rsproc->V();                  // Rendu processeur
  Sim()->Sched()->Schedule(55,Sim()->Tnow(),client);

}

// CLASSE BuffMgr : Evenement 5 - Transfert reseau

void BuffMgr::Event55(Client *client) {
  
  this->V();                    // Liberation RA5
  delete client->pgnums;        // Liberation memoire
  Sim()->Sched()->Schedule(43,Sim()->Tnow()+tnet,client);
  
}

void BuffMgr::RemoveNonRes(set<int> * oid_set) {
  set<int> temp_pset;
  set<int>::iterator it;
  int page_num;

  for (int i = 0; i < BUFFSIZE; i++) {
    if (buffer[i].pgnum != NULL_PAGE_NUM) {
      temp_pset.insert(buffer[i].pgnum);
    }
  }
  
  for (it = oid_set->begin(); it != oid_set->end(); it++) {
    page_num = DB->GetObj(*it)->PageDisk()->Scan(1);
    if (temp_pset.find(page_num) == temp_pset.end()) {
#ifdef DEBUG
      cout << "erasing obj" << endl;
#endif
      oid_set->erase(*it);
    } 
  }  
}


void BuffMgr::RemoveNonUpdated(set<int> * oid_set) {
  set<int> temp_pset;
  set<int>::iterator it;
  int page_num;

  for (int i = 0; i < BUFFSIZE; i++) {
    if ((buffer[i].pgnum != NULL_PAGE_NUM) && (buffer[i].updated == 1)) {
      temp_pset.insert(buffer[i].pgnum);
    }
  }
  
  for (it = oid_set->begin(); it != oid_set->end(); it++) {
    page_num = DB->GetObj(*it)->PageDisk()->Scan(1);
    if (temp_pset.find(page_num) == temp_pset.end()) {
#ifdef DEBUG
      cout << "erasing obj" << endl;
#endif
      oid_set->erase(*it);
    } 
  }  
}


// CLASSE BuffMgr : R4 - Acceder page(s)

// En sortie, la liste des pages a acceder ne contient plus que les
// Pages qui ne sont pas dans le cache (pages a acceder sur disque)
// This method deletes client->pgnums if the pages are in memory otherwise
// it leaves it for us to fetch from disk through AccDisk
// This method also updates the buffer replacement data.
void BuffMgr::AccPage(Client *client) {

  int i, j, k, n;
  short paginmem;

  i=1;
  n=client->pgnums->Scan(i);
  while (n!=-1) {
    j=0;
    paginmem=0;
    while ((j<BUFFSIZE) && (!paginmem))
      if (buffer[j].pgnum==n) {
        paginmem=1;
//        printf("Object #%d: page #%d in memory\n",client->rootoid,n);
      } else j++;
    if (paginmem) {
      // Page en memoire cache
      client->pgnums->Delete(i);        // Effacee de pgnums
      if (client->update_root_obj == true) {
	buffer[j].updated = 1;
      }
      // Mise en place flags remplacement
      buffer[j].nbacc++;
      if (KLRU>1)
        for (k=KLRU-1; k>=1; k--)
          buffer[j].date[k]=buffer[j].date[k-1];
      buffer[j].date[0]=Sim()->Tnow();
      buffer[j].clockflag=1;
      

      if (PGREP == 7) {
	int pgnum = n;
	float correl_period_of_refd_page;
#ifdef LATE_DEBUG
	cout << "updating pgnum = " << pgnum << endl;
#endif
	if ((Sim()->Tnow() - LRU_K_Last[pgnum]) > LRU_K_CORR_REF_PERIOD) {
	  correl_period_of_refd_page = 
	    LRU_K_Last[pgnum] - LRU_K_Hist[pgnum]->date[0];      
	  for (int index = 1; index < KLRU; index++) {
	    LRU_K_Hist[pgnum]->date[index] = 
	      LRU_K_Hist[pgnum]->date[index - 1] + correl_period_of_refd_page;
	  }
	  LRU_K_Hist[pgnum]->date[0] = Sim()->Tnow();
	  LRU_K_Last[pgnum] = Sim()->Tnow();
	}
	else {
	  LRU_K_Last[pgnum] = Sim()->Tnow();
	}
      }
    } else i++;
    n=client->pgnums->Scan(i);
  }
}

// CLASSE BuffMgr : Remplacement des pages dans le cache

int BuffMgr::Replace() {

  if (buffpos<BUFFSIZE-1) {     // Cache non plein, pas de remplacement
    buffpos++;
    return buffpos;
  } else {                      // Cache plein => page a remplacer
    switch (PGREP) {
      case 1: return RANDOMrep();break;
      case 2: return FIFOrep();break;
      case 3: return LFUrep();break;
      case 4: return LRUrep();break;
      case 5: return CLOCKrep();break;
      case 6: return GCLOCKrep();break;
      case 7: return LRUHistRep();break;
      case 8: return Otherep();break;
      default: printf("\nError: (%d) unknown page replacement policy!\n",PGREP);
    }
  }
  cerr << "ERROR:eventm.h reached here without returning" << endl;
  return 0;
}



// CLASSE BuffMgr : Remplacement des pages dans le cache - RANDOM

int BuffMgr::RANDOMrep() {

  // return CRand(1,0,BUFFSIZE-1);
  // using rand() function instead of CRand() since CRand is the random
  // function used by the rest of the system including workload generation.
  // Sometimes we need to have the workload generated twice, identically.
  // Using a different random function makes the identical generation
  // possible.
  return (int)(((float) rand() / (float) RAND_MAX) * ((float) (BUFFSIZE -1)));
}

// CLASSE BuffMgr : Remplacement des pages dans le cache - FIFO

int BuffMgr::FIFOrep() {

  BuffSlot temp;
  int i, j;

  // Decalage des places du cache d'un rang vers la "gauche". La place
  // en premiere position est placee en derniere position et sera
  // remplacee.

  temp.pgnum=buffer[0].pgnum;
  temp.updated=buffer[0].updated;
  temp.nbacc=buffer[0].nbacc;
  for (j=0; j<KLRU; j++) temp.date[j]=buffer[0].date[j];
  temp.clockflag=buffer[0].clockflag;

  for (i=0; i<BUFFSIZE-1; i++) {
    buffer[i].pgnum=buffer[i+1].pgnum;
    buffer[i].updated=buffer[i+1].updated;
    buffer[i].nbacc=buffer[i+1].nbacc;
    for (j=0; j<KLRU; j++) buffer[i].date[j]=buffer[i+1].date[j];
    buffer[i].clockflag=buffer[i+1].clockflag;
  }

  buffer[BUFFSIZE-1].pgnum=temp.pgnum;
  buffer[BUFFSIZE-1].updated=temp.updated;
  buffer[BUFFSIZE-1].nbacc=temp.nbacc;
  for (j=0; j<KLRU; j++) buffer[BUFFSIZE-1].date[j]=temp.date[j];
  buffer[BUFFSIZE-1].clockflag=temp.clockflag;

  return (BUFFSIZE-1);
}

// CLASSE BuffMgr : Remplacement des pages dans le cache - LFU

int BuffMgr::LFUrep() {

  int i, min, p;

  // Parcours du cache pour trouver la page avec le nombre d'acces
  // le plus faible.

  min=buffer[clock].nbacc;
  p=clock;
  i=clock+1;
  if (i>=BUFFSIZE) i=0;
  while (i!=clock) {
    if (buffer[i].nbacc<min) {
      min=buffer[i].nbacc;
      p=i;
    }
    i++;
    if (i>=BUFFSIZE) i=0;
  }
  clock=p+1;
  if (clock>=BUFFSIZE) clock=0;
  return p;
}

// CLASSE BuffMgr : Remplacement des pages dans le cache - LRU-K

int BuffMgr::LRUrep() {

  int i, p, j;
  float dist, max;
  // Parcours du cache pour trouver la page avec la distance
  // d'utilisation la plus forte
  
  max=0;
  for (j=0; j<KLRU; j++) max+=Sim()->Tnow()-buffer[clock].date[j];
  p=clock;
  i=clock+1;
  if (i>=BUFFSIZE) i=0;
  while (i!=clock) {
    dist=0;
    for (j=0; j<KLRU; j++) dist+=Sim()->Tnow()-buffer[i].date[j];
    if (dist>max) {
      max=dist;
      p=i;
    }
    i++;
    if (i>=BUFFSIZE) i=0;
  }
  clock=p+1;
  if (clock>=BUFFSIZE) clock=0;
  return p;
}

int BuffMgr::LRUHistRep() {
  
  int i, p;
  float min = Sim()->Tnow();

  p=clock;
  i=clock+1;
  if (i>=BUFFSIZE) i=0;
  while (i!=clock) {
    int tempPgNb = buffer[i].pgnum;
    
      if (((Sim()->Tnow() - LRU_K_Last[tempPgNb]) > LRU_K_CORR_REF_PERIOD) && 
	  ((LRU_K_Hist[tempPgNb]->date[KLRU-1]) < min)) {
	
	p = i;
	min = LRU_K_Hist[tempPgNb]->date[KLRU-1];
	
      }
      
      i++;
      if (i>=BUFFSIZE) i=0;
  }
  clock=p+1;
  if (clock>=BUFFSIZE) clock=0;
  return p;
}


void BuffMgr::DisplayHeat() {
  int heat;
  for (int i = 0; i < BUFFSIZE; i++) {
    if (DB->GetPage(buffer[i].pgnum) != NULL) {
      heat = DB->GetPage(buffer[i].pgnum)->CalcHeat();
      cout << "page  : " << buffer[i].pgnum<< endl;
      cout << "heat  : " << heat << endl;
    }
  }
}

void BuffMgr::StartIter(int * iter) {
  *iter =  0;
}

int BuffMgr::NextPgNum(int * iter) {
  int r_value;
  
  if (*iter < BUFFSIZE) {
    r_value = buffer[*iter].pgnum; 
    if (r_value != NULL_PAGE_NUM) {
      (*iter)++;
    }
    else {
      while ((r_value == NULL_PAGE_NUM) && (*iter < BUFFSIZE)) {
	r_value = buffer[*iter].pgnum;	  
	(*iter)++;	
      }
    }
  }
  else {
    r_value = NULL_PAGE_NUM;
  }
  return r_value;
}


int BuffMgr::NextNonEmptyPgNum(int * iter) {
  int r_value;
  
  if (*iter < BUFFSIZE) {
    r_value = buffer[*iter].pgnum; 
    if ((r_value != NULL_PAGE_NUM) && (DB->PageEmpty(r_value) == false)) {
      (*iter)++;
    }
    else {
      while (((r_value == NULL_PAGE_NUM) || (DB->PageEmpty(r_value) == true)) 
	     && (*iter < BUFFSIZE)) {
	r_value = buffer[*iter].pgnum;	  
	(*iter)++;	
      }
    }
  }
  else {
    r_value = NULL_PAGE_NUM;
  }
  return r_value;
}


int BuffMgr::NextUpdatedPgNum(int * iter) {
  int r_value;
  int page_updated;

  if (*iter < BUFFSIZE) {
    r_value = buffer[*iter].pgnum;
    page_updated = buffer[*iter].updated;
    if ((r_value != NULL_PAGE_NUM) && (buffer[*iter].updated == 1)) {
      (*iter)++;
    }
    else {      
      while (((r_value == NULL_PAGE_NUM) || (page_updated == 0))
	     && (*iter < BUFFSIZE)) {
	r_value = buffer[*iter].pgnum;
	page_updated = buffer[*iter].updated;
	(*iter)++;	
      }
    }
  }
  else {
    r_value = NULL_PAGE_NUM;
  }
  return r_value;
}

int BuffMgr::NextNonEmptyUpdatedPgNum(int * iter) {
  int r_value;
  int page_updated;

  if (*iter < BUFFSIZE) {
    r_value = buffer[*iter].pgnum;
    page_updated = buffer[*iter].updated;
    if ((r_value != NULL_PAGE_NUM) && (buffer[*iter].updated == 1)
	&& (DB->PageEmpty(r_value) == false)) {
      (*iter)++;
    }
    else {      
      while (((r_value == NULL_PAGE_NUM) || (page_updated == 0) ||
	      (DB->PageEmpty(r_value) == true))
	     && (*iter < BUFFSIZE)) {
	r_value = buffer[*iter].pgnum;
	page_updated = buffer[*iter].updated;
	(*iter)++;	
      }
    }
  }
  else {
    r_value = NULL_PAGE_NUM;
  }
  return r_value;
}



void BuffMgr::DisplayExtTen() {
  int et;

  for (int i = 0; i < BUFFSIZE; i++) {
    if (DB->GetPage(buffer[i].pgnum) != NULL) {
      et = DB->GetPage(buffer[i].pgnum)->CalcExtTension(this);
      cout << "page    : " << buffer[i].pgnum<< endl;
      cout << "tension : " << et << endl;
    }
  }
}

// CLASSE BuffMgr : Remplacement des pages dans le cache - CLOCK

int BuffMgr::CLOCKrep() {

  // Parcours du cache jusqu'a trouver une page dont le flag
  // est a 0 (page non utilisee). Au passage, les flags 1 sont
  // remis a 0.

  while (buffer[clock].clockflag) {
    buffer[clock].clockflag=0;
    clock++;
    if (clock>=BUFFSIZE) clock=0;
  }
  return clock;
}

// CLASSE BuffMgr : Rmplacement des pages dans le cache - GCLOCK

int BuffMgr::GCLOCKrep() {

  // Parcours du cache jusqu'a trouver une page dont le nombre
  // d'acces est 0. Decrementation du nombre d'acces, au passage.

  while (buffer[clock].nbacc) {
    buffer[clock].nbacc--;
    clock++;
    if (clock>=BUFFSIZE) clock=0;
  }
  return clock;
}

// CLASSE BuffMgr : Remplacement des pages dans le cache - Politique utilisateur

int BuffMgr::Otherep() {

  // Code a inserer ici
  // need to add code here
  return 1;
}

/////////////////////////////////////////////////////////////////////
// CLASSE IOSubSys (RA6)
/////////////////////////////////////////////////////////////////////

// CLASSE IOSubSys : Constructeur

IOSubSys::IOSubSys(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsd): Resource(n, cap, sim) {

  rsproc=rsp;
  rsdisk=rsd;
  lastpage=-999;
}

// CLASSE IOSubSys : Evenement 0 - Reservation RA6

void IOSubSys::Event60(Client *client) {

  this->P(61,client,1);
}

void IOSubSys::SetBuffPtr(BuffMgr * bm) {
  buffmgr = bm;
}

// CLASSE IOSubSys : Evenement 1 - Reservation processeur

void IOSubSys::Event61(Client *client) {

  rsproc->P(62,client,1);
}

// CLASSE IOSubSys : Evenement 2 - Reservation disque

void IOSubSys::Event62(Client *client) {

  client->pageindex=1;          // 1ere page a acceder
  rsdisk->P(63,client,1);
}

// CLASSE IOSubSys : Evenement 3 - Acces disque

void IOSubSys::Event63(Client *client) {
 int pagetoaccess;

  if (client->pageindex<=client->pgnums->NbElem()) {

    pagetoaccess=client->pgnums->Scan(client->pageindex);
    if (CLUSTP == 3) {
      DROPageStatsMap[pagetoaccess]->incNbLoad();
    }

    AccDisk(client);                    // R5
    // Stats
    NBIO++;
    if (client->type <= NTTRANS) {
      TNBIO++;
    }

  } else { // Fin
    rsdisk->V();                        // Rendu disque
    rsproc->V();                        // Rendu processeur
    this->V();                          // Liberation RA5
    if (client->type!=DDWT) // Operation de lecture => page en memoire
      Sim()->Sched()->Schedule(53,Sim()->Tnow(),client);
    else // Operation d'ecriture => fin de traitement, destruction du client
      Sim()->KillClient(client);
  }
}

// CLASSE IOSubSys : Acces disque (R5)

void IOSubSys::AccDisk(Client *client) {


  int pagetoaccess;

  pagetoaccess=client->pgnums->Scan(client->pageindex++);

  if ((pagetoaccess<(lastpage-1)) || (pagetoaccess>(lastpage+1))) {
    lastpage=pagetoaccess;
    // Non-contiguous access
    Sim()->Sched()->Schedule(63,Sim()->Tnow()+DISKSEA+DISKLAT+DISKTRA,client);
  } else {
    lastpage=pagetoaccess;
    // Contiguous access: no search time
    Sim()->Sched()->Schedule(63,Sim()->Tnow()+DISKLAT+DISKTRA,client);
  }
}
 
