
/////////////////////////////////////////////////////////////////////
// Package SIMULation (C++ discrete event simulation package)
// Version 1.1 g++, fevrier 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// Updated
// -------
// Name: Zhen He
// Institution: Australian National University
// Date: Feb. 2002
/////////////////////////////////////////////////////////////////////
// simulc.h : Definition des classes du moteur de simulation
// Invariable
/////////////////////////////////////////////////////////////////////

class Simulation;
class Scheduler;
class SchedulerCell;
class Resource;
class QueueCell;

class EventManager; // Definie dans le module variable eventc.hh
class Client;       // Definie dans le module variable eventc.hh

/////////////////////////////////////////////////////////////////////
// Constantes
/////////////////////////////////////////////////////////////////////

#define STRS 25               // Taille des noms de ressources

/////////////////////////////////////////////////////////////////////
// CLASSE Simulation
/////////////////////////////////////////////////////////////////////
// Controle de la simulation
/////////////////////////////////////////////////////////////////////

class Simulation {

  public:

    // Methodes

    Simulation(float start, float max, long int seed); // Constructeur
    ~Simulation();                      // Destructeur
    
    // Execution d'une simulation
    void Run(int nreplic); 

    Scheduler *Sched();                      // Retourne l'adresse de l'ech.
    float Tnow();                            // Retourne la valeur de tnow
    float Tmax();                            // Retourne la valeur de tmax
    void Reset(float start, float max, long int seed); // Reinitialisation
    Client *NewClient();                // Creation client dans clientlist
    void KillClient(Client *client);    // Suppression client dans clientlist
    void PurgeClientList();             // Suppression de tous les clients

  private:

    // Attributs prives

    float tstart;                       // Debut de simulation
    float tmax;                         // Fin de simulation
    float tnow;                         // Date courante
    long int rseed;                     // Germe du generateur aleatoire
    Client *clientlist;                 // Liste des clients
    Scheduler *scheduler;               // Pointeur vers l'echeancier
    EventManager *eventmanager;         // Pointeur vers le gest. d'event.

};

/////////////////////////////////////////////////////////////////////
// CLASSE Scheduler
/////////////////////////////////////////////////////////////////////
// Echeancier (liste triee des evenements)
/////////////////////////////////////////////////////////////////////

class Scheduler {

  public:

    // Methodes

    Scheduler();                        // Constructeur
    ~Scheduler();                       // Destructeur
    int IsEmpty();                      // Retourne l'etat de l'echeancier
    void Schedule(int eventcode, float eventdate, Client *client); // Insertion
    int GetEventCode();                 // Retourne le code du 1er evnt
    float GetEventDate();               // Retourne la date du 1er evnt
    Client *GetClient();                // Retourne le client a servir
    void DestroyEvent();                // Detruit le 1er evenement
    void Purge();                       // Vide l'echeancier

  private:

    // Attributs prives

    SchedulerCell *top;                 // Pointeur vers le 1er evenement
    SchedulerCell *bottom;              // Pointeur vers le dernier evnt

};

/////////////////////////////////////////////////////////////////////
// CLASSE SchedulerCell
/////////////////////////////////////////////////////////////////////
// Cellule de l'echeancier
/////////////////////////////////////////////////////////////////////

class SchedulerCell {

  public:

    // Methodes

    SchedulerCell(int code, float date, Client *cli); // Constructeur
    int Code();                         // Retourne le code evnt
    float Date();                       // Retourne la date evnt
    Client *Cli();                      // Retourne le client traite
    SchedulerCell *Next();              // Retourne la suivante
    SchedulerCell *Previous();          // Retourne la precedente
    void SetNext(SchedulerCell *newnext); // Nouvelle suivante
    void SetPrevious(SchedulerCell *newprev); // Nouvelle precedente

  private:

    // Attributs prives

    int eventcode;                      // Code de l'evenement
    float eventdate;                    // Date de l'evenement
    Client *client;                     // Client traite
    SchedulerCell *next;                // Cellule suivante
    SchedulerCell *previous;            // Cellule precedente

};

/////////////////////////////////////////////////////////////////////
// CLASSE Resource
/////////////////////////////////////////////////////////////////////
// Ressource passive (les ressources actives sont une sous-classe)
/////////////////////////////////////////////////////////////////////

class Resource {

  public:

    // Methodes publiques

    Resource(char n[STRS], int cap, Simulation *sim); // Constructeur
    ~Resource();                        // Destructeur
    void PurgeQueue();                  // Vide la file d'attente
    void P(int event, Client *client, int prior); // Reservation ressource
    void V();                           // Liberation ressource
    Simulation *Sim();                  // Retourne l'adr. de l'obj. simul.
    void ResetCounters();               // Reinitialisation des compteurs
    void ResetStats();                  // Reinitialisation stats globales
    void Stats();                       // Calcul des stats
    void DisplayStats();                // Affichage des stats
    float Mean(short i);                // Retourne les stats (moyenne)
    float Dev(short i);                 // Retourne les stats (ecart-type)

  private:

    // Methodes internes

    void EnQueue(int eventcode, Client *client, int priority); // Insertion
    int GetEventCode();                 // Retourne le code du 1er en file
    Client *GetClient();                // Retourne le 1er client en file
    void DestroyTop();                  // Detruit le 1er elt de la file
    int QueueEmpty();                   // Etat de la file d'attente

    // Attributs prives

    char name[STRS];                    // Nom de la ressource
    QueueCell *top;                     // Tete de file d'attente
    QueueCell *bottom;                  // Fin de file d'attente
    int capacity;                       // Capacite ressource
    int ccapacity;                      // Capacite courante
    Simulation *simul;                  // Ptr sur l'objet simulation
    float response;                     // Temps de reponse sur une rep
    float wait;                         // Temps d'attente sur une rep
    int nbserv;                         // Nombre de clients servis
    float stats[5],stats2[5];           // Stats (cumul)
    int n;                              // Stats (nombre d'experiences)
    float mean[5], dev[5], cint[5];     // Moyennes - ecarts-types - intervalles de confiance
                                        // 0 : Temps de reponse
                                        // 1 : Temps d'attente
                                        // 2 : Nombre de clients servis
                                        // 3 : Nombre de clients en cours
                                        // 4 : Nombre de clients en attente

};

/////////////////////////////////////////////////////////////////////
// CLASSE QueueCell
/////////////////////////////////////////////////////////////////////
// Cellule de file d'attente ressource
/////////////////////////////////////////////////////////////////////

class QueueCell {

  public:

    // Methodes

    QueueCell(int code, Client *cli, int prior); // Constructeur
    int Code();                         // Retourne le code evnt
    Client *Cli();                      // Retourne le client
    int Priority();                     // Retourne la priorite
    QueueCell *Next();                  // Retourne la suivante
    QueueCell *Previous();              // Retourne la precedente
    void SetNext(QueueCell *newnext);   // Nouvelle suivante
    void SetPrevious(QueueCell *newprev); // Nouvelle precedente

  private:

    // Attributs prives

    int eventcode;                      // Code evenement
    Client *client;                     // Client
    int priority;                       // Priorite
    QueueCell *next;                    // Cellule suivante
    QueueCell *previous;                // Cellule precedente

};
